from ._mapi import *

class CS:

    CSA = []  

    def __init__(self, 
                 name: str,
                 duration: float = 0, 
                 s_group: str = None, 
                 s_age: float = None, 
                 s_type: str= None, 
                 b_group: str = None, 
                 b_pos: str = None, 
                 b_type: str = None,
                 l_group: str = None, 
                 l_day: str = None, 
                 l_type: str = None, 
                 id: int = None, 
                 sr_stage: bool = True, 
                 ad_stage: bool = False, 
                 load_in: bool = False, 
                 nl: int = 5, 
                 addstp: list = None):
        """
        Construction Stage define.
        
        Parameters:
            name: Name of Construction Stage
            duration: Duration of Construction Stage in days (default 0)
            s_group: Structure group name or list of group names (default None)
            s_age: Age of structure group in days and Redistribution value(%) win case of Deactivation (default 0)
            s_type: Structure activation type - "A" to activate, "D" to deactivate(default A)
            b_group: Boundary group name or list of group names (default None)
            b_pos: Boundary position type - "ORIGINAL" or "DEFORMED", or list (default DEFORMED)
            b_type: Boundary activation type - "A" to activate, "D" to deactivate (default A)
            l_group: Load group name or list of group names (default None)
            l_day: Load activation day - "FIRST" or "LAST" (default "FIRST")
            l_type: Load activation type - "A" to activate, "D" to deactivate (default A)
            id: The construction stage ID (optional)
            sr_stage: Save results of this stage (default True)
            ad_stage: Add additional step results (default False)
            load_in: Load incremental steps for material nonlinear analysis (default False)
            nl: Number of load incremental steps (default 5)
            addstp: List of additional steps (default None)
        
        Examples:
            ```python
            # Single group activation
            CS("CS1", 7, "S1", 7, "A", "B1", "DEFORMED", "A", "L1", "FIRST", "A")
            
            # Multiple group activation
            CS("CS1", 7, ["S1", "S2"], [7, 10], ["A", "A"], ["B1", "B2"], 
               ["DEFORMED", "DEFORMED"], ["A", "A"], ["L1", "L2"], ["FIRST", "FIRST"], ["A", "A"])
            
            # Mixed activation and deactivation
            CS("CS1", 7, ["S1", "S2"], [7, 10], ["A", "D"], ["B1", "B2"], 
               ["DEFORMED", "DEFORMED"], ["A", "D"], "L1", "FIRST", "A")
            
            # With additional options
            CS("CS1", 7, "S1", 7, "A", "B1", "DEFORMED", "A", "L1", "FIRST", "A",
               sr_stage=True, ad_stage=True, load_in=True, nl=6, addstp=[1, 2, 3])
            ```
        """

        self.NAME = name
        self.DURATION = duration
        self.SR_stage = sr_stage
        self.Ad_stage = ad_stage
        self.Load_IN = load_in
        self.NL = nl
        self.addstp = [] if addstp is None else addstp
        
        # Initialize group containers
        self.act_structure_groups = []  
        self.deact_structure_groups = []  
        self.act_boundary_groups = []  
        self.deact_boundary_groups = []  
        self.act_load_groups = []  
        self.deact_load_groups = []  
        
        # Set ID
        if id is None:
            self.ID = len(CS.CSA) + 1
        else:
            self.ID = id
        
        # Process structure groups
        if s_group:
            # Convert single values to lists for uniform processing
            if not isinstance(s_group, list):
                s_group = [s_group]
                s_age = [s_age if s_age is not None else 0]
                s_type = [s_type if s_type is not None else "A"]
            else:
                # Ensure other parameters are lists too
                if s_age is None:
                    s_age = [0] * len(s_group)
                elif not isinstance(s_age, list):
                    s_age = [s_age] * len(s_group)
                
                if s_type is None:
                    s_type = ["A"] * len(s_group)
                elif not isinstance(s_type, list):
                    s_type = [s_type] * len(s_group)
            
            # Process each structure group
            for i, group in enumerate(s_group):
                if i < len(s_type) and s_type[i] == "A":
                    # Activation: Check if already activated in previous stages
                    for stage in CS.CSA:
                        for existing_group in stage.act_structure_groups:
                            if existing_group["name"] == group:
                                raise ValueError(f"Structure group '{group}' has already been activated in stage '{stage.NAME}' (ID: {stage.ID})")
                    
                    age = s_age[i] if i < len(s_age) else 0
                    self.act_structure_groups.append({"name": group, "age": age})
                else:
                    # Deactivation: Check if activated in previous stages
                    activated = False
                    for stage in CS.CSA:
                        for existing_group in stage.act_structure_groups:
                            if existing_group["name"] == group:
                                activated = True
                                break
                        if activated:
                            break
                    
                    if not activated:
                        raise ValueError(f"Structure group '{group}' cannot be deactivated as it has not been activated in any previous stage")
                    
                    # For deactivation, s_age value is used as redist percentage
                    redist = s_age[i] if i < len(s_age) else 100
                    self.deact_structure_groups.append({"name": group, "redist": redist})
        
        # Process boundary groups
        if b_group:
            # Convert single values to lists for uniform processing
            if not isinstance(b_group, list):
                b_group = [b_group]
                b_pos = [b_pos if b_pos is not None else "DEFORMED"]
                b_type = [b_type if b_type is not None else "A"]
            else:
                # Ensure other parameters are lists too
                if b_pos is None:
                    b_pos = ["DEFORMED"] * len(b_group)
                elif not isinstance(b_pos, list):
                    b_pos = [b_pos] * len(b_group)
                
                if b_type is None:
                    b_type = ["A"] * len(b_group)
                elif not isinstance(b_type, list):
                    b_type = [b_type] * len(b_group)
            
            # Process each boundary group
            for i, group in enumerate(b_group):
                if i < len(b_type) and b_type[i] == "A":
                    # Activation: Check if already activated in previous stages
                    for stage in CS.CSA:
                        for existing_group in stage.act_boundary_groups:
                            if existing_group["name"] == group:
                                raise ValueError(f"Boundary group '{group}' has already been activated in stage '{stage.NAME}' (ID: {stage.ID})")
                    
                    pos = b_pos[i] if i < len(b_pos) else "DEFORMED"
                    self.act_boundary_groups.append({"name": group, "pos": pos})
                else:
                    # Deactivation: Check if activated in previous stages
                    activated = False
                    for stage in CS.CSA:
                        for existing_group in stage.act_boundary_groups:
                            if existing_group["name"] == group:
                                activated = True
                                break
                        if activated:
                            break
                    
                    if not activated:
                        raise ValueError(f"Boundary group '{group}' cannot be deactivated as it has not been activated in any previous stage")
                    
                    self.deact_boundary_groups.append(group)
        
        # Process load groups
        if l_group:
            # Convert single values to lists for uniform processing
            if not isinstance(l_group, list):
                l_group = [l_group]
                l_day = [l_day if l_day is not None else "FIRST"]
                l_type = [l_type if l_type is not None else "A"]
            else:
                # Ensure other parameters are lists too
                if l_day is None:
                    l_day = ["FIRST"] * len(l_group)
                elif not isinstance(l_day, list):
                    l_day = [l_day] * len(l_group)
                
                if l_type is None:
                    l_type = ["A"] * len(l_group)
                elif not isinstance(l_type, list):
                    l_type = [l_type] * len(l_group)
            
            # Process each load group
            for i, group in enumerate(l_group):
                if i < len(l_type) and l_type[i] == "A":
                    # Activation: Check if already activated in previous stages
                    for stage in CS.CSA:
                        for existing_group in stage.act_load_groups:
                            if existing_group["name"] == group:
                                raise ValueError(f"Load group '{group}' has already been activated in stage '{stage.NAME}' (ID: {stage.ID})")
                    
                    day = l_day[i] if i < len(l_day) else "FIRST"
                    self.act_load_groups.append({"name": group, "day": day})
                else:
                    # Deactivation: Check if activated in previous stages
                    activated = False
                    for stage in CS.CSA:
                        for existing_group in stage.act_load_groups:
                            if existing_group["name"] == group:
                                activated = True
                                break
                        if activated:
                            break
                    
                    if not activated:
                        raise ValueError(f"Load group '{group}' cannot be deactivated as it has not been activated in any previous stage")
                    
                    day = l_day[i] if i < len(l_day) else "FIRST"
                    self.deact_load_groups.append({"name": group, "day": day})
        
        CS.CSA.append(self)
    
    @classmethod
    def json(cls):
        """
        Converts Construction Stage data to JSON format 
        Example:
            # Get the JSON data for all construction stages
            json_data = CS.json()
            print(json_data)
        """
        json = {"Assign": {}}
        
        for csa in cls.CSA:
            stage_data = {
                "NAME": csa.NAME,
                "DURATION": csa.DURATION,
                "bSV_RSLT": csa.SR_stage,
                "bSV_STEP": csa.Ad_stage,
                "bLOAD_STEP": csa.Load_IN
            }
            
            # Add incremental steps if load step is enabled
            if csa.Load_IN:
                stage_data["INCRE_STEP"] = csa.NL
            
            # Add additional steps if specified
            if csa.addstp:
                stage_data["ADD_STEP"] = csa.addstp
            else:
                stage_data["ADD_STEP"] = []
            
            # Handle structure group activation
            if csa.act_structure_groups:
                stage_data["ACT_ELEM"] = []
                for group in csa.act_structure_groups:
                    stage_data["ACT_ELEM"].append({
                        "GRUP_NAME": group["name"],
                        "AGE": group["age"]
                    })
            
            # Handle structure group deactivation
            if csa.deact_structure_groups:
                stage_data["DACT_ELEM"] = []
                for group in csa.deact_structure_groups:
                    stage_data["DACT_ELEM"].append({
                        "GRUP_NAME": group["name"],
                        "REDIST": group["redist"]
                    })
            
            # Handle boundary group activation
            if csa.act_boundary_groups:
                stage_data["ACT_BNGR"] = []
                for group in csa.act_boundary_groups:
                    stage_data["ACT_BNGR"].append({
                        "BNGR_NAME": group["name"],
                        "POS": group["pos"]
                    })
            
            # Handle boundary group deactivation
            if csa.deact_boundary_groups:
                stage_data["DACT_BNGR"] = []
                for group_name in csa.deact_boundary_groups:
                    stage_data["DACT_BNGR"].append(group_name)
            
            # Handle load group activation
            if csa.act_load_groups:
                stage_data["ACT_LOAD"] = []
                for group in csa.act_load_groups:
                    stage_data["ACT_LOAD"].append({
                        "LOAD_NAME": group["name"],
                        "DAY": group["day"]
                    })
            
            # Handle load group deactivation
            if csa.deact_load_groups:
                stage_data["DACT_LOAD"] = []
                for group in csa.deact_load_groups:
                    stage_data["DACT_LOAD"].append({
                        "LOAD_NAME": group["name"],
                        "DAY": group["day"]
                    })
            
            json["Assign"][str(csa.ID)] = stage_data
        
        return json
    
    @classmethod
    def create(cls):
        """Creates construction stages in the database"""
        MidasAPI("PUT", "/db/stag", CS.json())
    
    @classmethod
    def get(cls):
        """Gets construction stage data from the database"""
        return MidasAPI("GET", "/db/stag")
    
    @classmethod
    def sync(cls):
        """Updates the CS class with data from the database"""
        cls.CSA = []
        a = CS.get()
        if a != {'message': ''}:
            if "STAG" in a:
                stag_data_dict = a["STAG"]
            else:
                return  
                
            for stag_id, stag_data in stag_data_dict.items():
                # Basic stage data
                name = stag_data.get("NAME")
                duration = stag_data.get("DURATION")
                sr_stage = stag_data.get("bSV_RSLT")
                ad_stage = stag_data.get("bSV_STEP")
                load_in = stag_data.get("bLOAD_STEP")
                nl = stag_data.get("INCRE_STEP")
                addstp = stag_data.get("ADD_STEP")
                
                # Create a new CS object with basic properties
                new_cs = CS(
                    name=name,
                    duration=duration,
                    id=int(stag_id),
                    sr_stage=sr_stage,
                    ad_stage=ad_stage,
                    load_in=load_in,
                    nl=nl,
                    addstp=addstp
                )
                
                CS.CSA.pop()
                
                # Process activation elements
                if "ACT_ELEM" in stag_data and stag_data["ACT_ELEM"]:
                    for elem in stag_data["ACT_ELEM"]:
                        group_name = elem.get("GRUP_NAME")
                        age = elem.get("AGE")
                        new_cs.act_structure_groups.append({"name": group_name, "age": age})
                
                # Process deactivation elements
                if "DACT_ELEM" in stag_data and stag_data["DACT_ELEM"]:
                    for elem in stag_data["DACT_ELEM"]:
                        if isinstance(elem, dict):
                            group_name = elem.get("GRUP_NAME")
                            redist = elem.get("REDIST")
                        else:
                            group_name = elem
                            redist = 0
                        new_cs.deact_structure_groups.append({"name": group_name, "redist": redist})
                
                # Process activation boundary groups
                if "ACT_BNGR" in stag_data and stag_data["ACT_BNGR"]:
                    for bngr in stag_data["ACT_BNGR"]:
                        group_name = bngr.get("BNGR_NAME")
                        pos = bngr.get("POS")
                        new_cs.act_boundary_groups.append({"name": group_name, "pos": pos})
                
                # Process deactivation boundary groups
                if "DACT_BNGR" in stag_data and stag_data["DACT_BNGR"]:
                    for bngr in stag_data["DACT_BNGR"]:
                        new_cs.deact_boundary_groups.append(bngr)
                
                # Process activation loads
                if "ACT_LOAD" in stag_data and stag_data["ACT_LOAD"]:
                    for load in stag_data["ACT_LOAD"]:
                        group_name = load.get("LOAD_NAME")
                        day = load.get("DAY")
                        new_cs.act_load_groups.append({"name": group_name, "day": day})
                
                # Process deactivation loads
                if "DACT_LOAD" in stag_data and stag_data["DACT_LOAD"]:
                    for load in stag_data["DACT_LOAD"]:
                        if isinstance(load, dict):
                            group_name = load.get("LOAD_NAME")
                            day = load.get("DAY")
                        else:
                            group_name = load
                            day = "FIRST"
                        new_cs.deact_load_groups.append({"name": group_name, "day": day})
                
                CS.CSA.append(new_cs)
    
    @classmethod
    def delete(cls):
        """Deletes all construction stages from the database and resets the class"""
        cls.CSA = []
        return MidasAPI("DELETE", "/db/stag")


import requests
import sys
from colorama import Fore, Style
try:import winreg
except: pass
import time
from tqdm import tqdm
# import polars as pl



def Midas_help():
    """MIDAS Documnetation : https://midas-rnd.github.io/midasapi-python """
    print('\n╭────────────────────────────────────────────────────────────────────────────────────╮')
    print("│         HELP MANUAL   :     https://midas-rnd.github.io/midasapi-python/           │")
    print('╰────────────────────────────────────────────────────────────────────────────────────╯\n')



class NX:
    version_check = True    # CHANGE IT TO FALSE TO SKIP VERSION CHECK OF LIBRARY
    user_print = True
    debug_request = False
    debug_requestJSON = False
    debug_response = False
    onlyNode = False
    visualiser = False

    

class MAPI_COUNTRY:
    
    country = "US"

    def __init__(self,country:str="US"):
        ''' Define Civil NX country to automatically set Base URL and MAPI Key from registry.
        ```
        MAPI_COUNTRY('US') # For english version
        MAPI_COUNTRY('KR') # For Korean version
        MAPI_COUNTRY('CH') # For Chinese version
        ```
        '''
        if country.lower() in ['us','ch','kr','jp']:
            MAPI_COUNTRY.country = country.upper()
        else:
            MAPI_COUNTRY.country = 'US'
        
        MAPI_BASEURL.setURLfromRegistry()
        MAPI_KEY.get_key()  # Intial Key from registry


class MAPI_BASEURL:
    baseURL = "https://moa-engineers.midasit.com:443/civil"
    server_loc = "Global"
    
    def __init__(self, baseURL:str = "https://moa-engineers.midasit.com:443/civil"):
        ''' Define the Base URL for API connection.
        ```
        MAPI_BASEURL('https://moa-engineers.midasit.com:443/civil')
        ```
        '''
        MAPI_BASEURL.baseURL = baseURL
        
    @classmethod
    def get_url(cls):
        return MAPI_BASEURL.baseURL
    
    @classmethod
    def setURLfromRegistry(cls):
        try:
            key_path = f"Software\\MIDAS\\CVLwNX_{MAPI_COUNTRY.country}\\CONNECTION"  
            registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, key_path, 0, winreg.KEY_READ)
            url_reg = winreg.QueryValueEx(registry_key, "URI")
            url_reg_key = url_reg[0]

            port_reg = winreg.QueryValueEx(registry_key, "PORT")
            port_reg_key = port_reg[0]

            url_comb = f'https://{url_reg_key}:{port_reg_key}/civil'

            tqdm.write(f' 🌐   BASE URL is taken from Registry entry.  >>  {url_comb}')
            MAPI_BASEURL(url_comb)
        except:
            tqdm.write(" 🌐   BASE URL is not defined. Click on Apps > API Settings to copy the BASE URL Key.\nDefine it using MAPI_BASEURL('https://moa-engineers.midasit.com:443/civil')")
            sys.exit(0)

    @staticmethod
    def autoURL():
        base_urls = [
            "https://moa-engineers-in.midasit.com:443/civil",
            "https://moa-engineers-kr.midasit.com:443/civil",
            "https://moa-engineers-gb.midasit.com:443/civil",
            "https://moa-engineers-us.midasit.com:443/civil",
            "https://moa-engineers.midasit.cn:443/civil"
            ]
        serv_locations = ["INDIA","KOREA","EUROPE","USA","CHINA"]
        mapi_key = MAPI_KEY.get_key()
        chk = 0
        for i,base_url in enumerate(base_urls):
            url = base_url + "/config/ver"
            headers = {
                "Content-Type": "application/json",
                "MAPI-Key": mapi_key
            }
            response = requests.get(url=url, headers=headers)
            if response.status_code == 200:
                MAPI_BASEURL(base_url)
                MAPI_BASEURL.server_loc = serv_locations[i]
                chk=1
                break
        if chk==0:
            tqdm.write(f" 🌐   Kindly manually enter the BASE URL. \nRefer to https://moa.midasit.com/services to find the correct URL.")
            sys.exit(0)
            
class MAPI_KEY:
    """MAPI key from Civil NX.\n\nEg: MAPI_Key("eadsfjaks568wqehhf.ajkgj345qfhh")"""
    data = ""
    count = 1
    
    def __init__(self, mapi_key:str):
        MAPI_KEY.data = mapi_key
        
    @classmethod
    def get_key(cls):
        if MAPI_KEY.data == "":
            try:
                key_path = f"Software\\MIDAS\\CVLwNX_{MAPI_COUNTRY.country}\\CONNECTION"  
                registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, key_path, 0, winreg.KEY_READ)
                value = winreg.QueryValueEx(registry_key, "Key")
                my_key = value[0]
                MAPI_KEY(my_key)
                tqdm.write(f' 🔑   MAPI-KEY is taken from Registry entry.  >>  {my_key[:35]}...')
            except:
                tqdm.write(f"🔑   MAPI KEY is not defined. Click on Apps > API Settings to copy the MAPI Key.\n Define it using MAPI_KEY('xxxx')")
                sys.exit(0)
        else:
            my_key = MAPI_KEY.data
        
        return my_key
#---------------------------------------------------------------------------------------------------------------

#2 midas API link code:
def MidasAPI(method:str, command:str, body:dict={})->dict:
    """Sends HTTP Request to MIDAS Civil NX
            Parameters:
                Method: "PUT" , "POST" , "GET" or "DELETE"
                Command: eg. "/db/NODE"
                Body: {{"Assign":{{1{{'X':0, 'Y':0, 'Z':0}}}}}}            
            Examples:
                ```python
                # Create a node
                MidasAPI("PUT","/db/NODE",{{"Assign":{{"1":{{'X':0, 'Y':0, 'Z':0}}}}}})"""
    
    base_url = MAPI_BASEURL.baseURL
    mapi_key = MAPI_KEY.get_key()

    url = base_url + command
    headers = {
        "Content-Type": "application/json",
        "MAPI-Key": mapi_key
    }

    if MAPI_KEY.count == 1:
        MAPI_KEY.count =0
        if NX.user_print:
            _checkUSER()




    start_time = time.perf_counter()


    if method == "POST":
        response = requests.post(url=url, headers=headers, json=body)
    elif method == "PUT":
        response = requests.put(url=url, headers=headers, json=body)
    elif method == "GET":
        response = requests.get(url=url, headers=headers)
    elif method == "DELETE":
        response = requests.delete(url=url, headers=headers)

    end_time = time.perf_counter()
    elapsed_time = end_time - start_time

    if NX.debug_request:
        tqdm.write(Fore.RED+f">>   METHOD : {method} |  URL : {command} | STATUS :  {response.status_code} | TIME : {elapsed_time:.4f} sec "+Style.RESET_ALL)
    if NX.debug_requestJSON:
        tqdm.write(Fore.CYAN+">>  "+str(body)+Style.RESET_ALL)
    if NX.debug_response:
        tqdm.write(Fore.GREEN+"<<  "+str(response.json())+Style.RESET_ALL)

    if MAPI_KEY.count == 0:
        MAPI_KEY.count = -1
        if response.status_code == 404:
            print(Fore.RED +'\n╭─ 💀   ─────────────────────────────────────────────────────────────────────────────╮')
            print(f"│  Civil NX model is not connected.  Click on 'Apps > Connect' in Civil NX.          │")
            print(f"│  Make sure the MAPI Key in python code is matching with the MAPI key in Civil NX.  │")
            print('╰────────────────────────────────────────────────────────────────────────────────────╯\n'+Style.RESET_ALL)
            sys.exit(0)
    


    return response.json()


#--------------------------------------------------------------------

def _getUNIT():
    return MidasAPI('GET','/db/UNIT',{})['UNIT']['1']

def _setUNIT(unitJS):
    js = {
        "Assign" : {
            "1" : unitJS
        }
    }
    MidasAPI('PUT','/db/UNIT',js)


def _checkUSER():
    try:
        resp =  MidasAPI('GET','/config/ver',{})['VER']

        # print(f"{' '*15}Connected to {resp['NAME']}")
        # print(f"{' '*15}USER : {resp['USER']}          COMPANY : {resp['COMPANY']}")

        ln1 = f"Connected to {resp['NAME']}            SERVER : {MAPI_BASEURL.server_loc}"
        ln2 = f"USER : {resp['USER']}          COMPANY : {resp['COMPANY']}"

        lg_ln1 = 66-len(ln1)
        lg_ln2 = 70-len(ln2)

        line1 = f"│{' '*12} {ln1} {' '*lg_ln1} 🟢 │"
        line2 = f"│{' '*12} {ln2} {' '*lg_ln2}│"
        tqdm.write(Fore.GREEN+'\n╭─ 🔔  ──────────────────────────────────────────────────────────────────────────────╮')
        tqdm.write(line1)
        tqdm.write(line2)
        tqdm.write('╰────────────────────────────────────────────────────────────────────────────────────╯\n'+Style.RESET_ALL)


        # print('─'*86)

    except:
        pass



from ._mapi import MidasAPI
# from ._utils import *
from ._load import Load_Case
from ._group import Group


def _JStoObj_Relax(js):
    rm = js['RM']
    rv = js['RV']
    us = js['US']
    ys = js['YS']
    bRelax = js['bRELAX']
    cff = js['FF']

    if rm == 9: # CEBFIP 2010 Code
        wtype = js['W_TYPE']
        if wtype:
            return Tendon.Relaxation.CEBFIP_2010(rv,js['TDMFK'],us,ys,cff,0,js['W_ANGLE'])
        else:
            return Tendon.Relaxation.CEBFIP_2010(rv,js['TDMFK'],us,ys,cff,js['WF'],0)
        
    elif rm == 8: # CEBFIP 1990 Code
        wtype = js['W_TYPE']
        if wtype:
            return Tendon.Relaxation.CEBFIP_1990(rv,us,ys,cff,0,js['W_ANGLE'])
        else:
            return Tendon.Relaxation.CEBFIP_1990(rv,us,ys,cff,js['WF'],0)
        
    elif rm == 1: # CEBFIP 2010 Code
        wtype = js['W_TYPE']
        if wtype:
            return Tendon.Relaxation.CEBFIP_1978(rv,us,ys,cff,0,js['W_ANGLE'])
        else:
            return Tendon.Relaxation.CEBFIP_1978(rv,us,ys,cff,js['WF'],0)
        
    elif rm == 5: # European
        wtype = js['W_TYPE']
        if wtype:
            return Tendon.Relaxation.European(rv,us,ys,cff,0,js['W_ANGLE'])
        else:
            return Tendon.Relaxation.European(rv,us,ys,cff,js['WF'],0)
    
    elif rm == 4:
        return Tendon.Relaxation.IRC_18(rv,us,ys,cff,js['WF'])
    
    elif rm == 7:
        return Tendon.Relaxation.IRC_112(rv,us,ys,cff,js['WF'])
    
    elif rm == 0:
        if bRelax:
            return Tendon.Relaxation.Magura(rv,us,ys,cff,js['WF'])
        else: return Tendon.Relaxation.Null(us,ys,cff,js['WF'])

    else : # If nothing is matched No relaxation is applied
        return Tendon.Relaxation.Null(us,ys,cff,js['WF'])

def _JStoObj_Prop(id, js):
    name = js['NAME']

    if js['TYPE'] == 'INTERNAL' :
        if js['LT'] == 'POST': type = 2
        else : type = 1
    else: type = 3


    matID = js['MATL']
    tdn_area = js['AREA']
    duct_dia = js['D_AREA']
    relax = _JStoObj_Relax(js)
    ext_mo = js['ALPHA']
    an_s_b = js['ASB']
    an_s_e = js['ASE']
    bond_type = js['bBONDED']
    
    Tendon.Property(name,type,matID,tdn_area,duct_dia,relax,ext_mo,an_s_b,an_s_e,bond_type,id)

def _JStoObj_Profile(id,js):
    tdn_id = id 
    name = js['NAME']
    tdnProperty = js['TDN_PROP']
    tdn_group = js['TDN_GRUP']
    elem = js['ELEM']  

    inp_type = js['INPUT'] #3D
    curve_type = js['CURVE'] #SPLINE
    ref_axis = js['SHAPE'] #ELEMENT

    # Common
    st_len_begin = js['BELENG']
    st_len_end = js['ELENG']

    b_typical_tendon = js['bTP']
    n_typical_tendon = 0
    if b_typical_tendon:
        n_typical_tendon = js['CNT']

    trans_len_opt = js['LENG_OPT']
    trans_len_begin = js['BLEN']
    trans_len_end = js['ELEN']

    debon_len_begin = js['DeBondBLEN']
    debon_len_end = js['DeBondELEN']


    #Variable initialise to remove errors
    prof_xyz = []  
    prof_xy = []  
    prof_xz = []  
    prof_ins_point_end = 'END-I'  
    prof_ins_point_elem = 0  
    x_axis_dir_element = 'I-J'  
    x_axis_rot_ang = 0  
    projection = True  
    offset_y = 0  
    offset_z = 0  
    prof_ins_point = [0, 0, 0]  
    x_axis_dir_straight = 'X'  
    x_axis_dir_vec = [0, 0]  
    grad_rot_axis = 'X'  
    grad_rot_ang = 0  
    radius_cen = [0, 0]  
    offset = 0  
    dir = 'CW'      

    
    #3D - SPLINE - ROUND is bFIX , R is not considered
    if inp_type == '3D' :
        prof_xyz = []  
        json_profile_arr = js['PROF']

        for i in range(len(json_profile_arr)):
            prof_xyz.append([json_profile_arr[i]['PT'][0],json_profile_arr[i]['PT'][1],json_profile_arr[i]['PT'][2]])

    

    #2D
    elif inp_type == '2D' :
        prof_xy = []  
        prof_xz = []  
        json_profileY_arr = js['PROFY']
        json_profileZ_arr = js['PROFZ']

        for i in range(len(json_profileY_arr)):
            prof_xy.append([json_profileY_arr[i]['PT'][0],json_profileY_arr[i]['PT'][1]])
        
        for i in range(len(json_profileZ_arr)):
            prof_xz.append([json_profileZ_arr[i]['PT'][0],json_profileZ_arr[i]['PT'][1]])


    # ELEMENT
    if ref_axis == 'ELEMENT' :

        prof_ins_point_end = js['INS_PT'] 
        prof_ins_point_elem = js['INS_ELEM']
        x_axis_dir_element = js['AXIS_IJ'] 
        x_axis_rot_ang =js['XAR_ANGLE']  
        projection = js['bPJ']
        offset_y = js['OFF_YZ'][0]
        offset_z = js['OFF_YZ'][1]

    elif ref_axis == 'STRAIGHT' :
        prof_ins_point = js['IP'] 
        x_axis_dir_straight = js['AXIS'] 
        x_axis_dir_vec = js['VEC'] 
        x_axis_rot_ang =js['XAR_ANGLE']  
        projection = js['bPJ']
        grad_rot_axis = js['GR_AXIS']  
        grad_rot_ang = js['GR_ANGLE'] 

    elif ref_axis == 'CURVE' :
        prof_ins_point = js['IP'] 
        radius_cen = js['RC']  
        offset = js['OFFSET']   
        dir = js['DIR'] 
        x_axis_rot_ang =js['XAR_ANGLE']  
        projection = js['bPJ']
        grad_rot_axis = js['GR_AXIS']  
        grad_rot_ang = js['GR_ANGLE']

    
    Tendon.Profile(name,tdnProperty,tdn_group,elem,inp_type,curve_type,st_len_begin,st_len_end,n_typical_tendon,
                   trans_len_opt,trans_len_begin,trans_len_end,debon_len_begin,debon_len_end,ref_axis,
                   prof_xyz,prof_xy,prof_xz,prof_ins_point,prof_ins_point_elem,x_axis_dir_element,x_axis_rot_ang,
                   projection,offset_y,offset_z,prof_ins_point,x_axis_dir_straight,x_axis_dir_vec,grad_rot_axis,grad_rot_ang,radius_cen,offset,dir,tdn_id)



class _POINT_ : # Local class to store points
    def __init__(self,x,y,z):
        self.X = x
        self.Y = y
        self.Z = z


#5 Class to create nodes
class Tendon:

    @staticmethod
    def create():
        if Tendon.Property.properties!=[]:
            Tendon.Property.create()
        if Tendon.Profile.profiles !=[]:
            Tendon.Profile.create()
        if Tendon.Prestress.loads !=[]:
            Tendon.Prestress.create()

    @staticmethod
    def clear():
        Tendon.Property.clear()
        Tendon.Profile.clear()
        Tendon.Prestress.clear()


    class Relaxation:

        class CEBFIP_2010:
            
            def __init__(self,rho,rel_class,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0,unint_ang_disp=0):
                '''
                    rel_class =  1 Slow | 2 Mean | 3 Rapid
                '''
                self.CODE = 'CEB FIP-2010'
                self.RHO = rho
                self.CLASS = rel_class

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac

                self.UNINT_AD = unint_ang_disp
            
            def json(self):
                bs_json ={
                    "RM" : 9,
                    "RV" : self.RHO,
                    "TDMFK" : self.CLASS,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF,
                }

                if self.UNINT_AD != 0:
                    bs_json.update({"W_TYPE" : 1 , "W_ANGLE" : self.UNINT_AD})

                return bs_json
            
            def __str__(self):
                return str(self.__dict__)
            
        class CEBFIP_1978:
            
            def __init__(self,rho,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0,unint_ang_disp=0):
                '''
                    rel_class =  1 Slow | 2 Mean | 3 Rapid
                '''
                self.CODE = 'CEB FIP-1978'
                self.RHO = rho

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac

                self.UNINT_AD = unint_ang_disp
            
            def json(self):
                bs_json ={
                    "RM" : 1,
                    "RV" : self.RHO,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF,
                }

                if self.UNINT_AD != 0:
                    bs_json.update({"W_TYPE" : 1 , "W_ANGLE" : self.UNINT_AD})

                return bs_json
            
            def __str__(self):
                return str(self.__dict__)
        
        class CEBFIP_1990:
            
            def __init__(self,rho,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0,unint_ang_disp=0):
                '''
                    rel_class =  1 Slow | 2 Mean | 3 Rapid
                '''
                self.CODE = 'CEB FIP-1990'
                self.RHO = rho

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac

                self.UNINT_AD = unint_ang_disp
            
            def json(self):
                bs_json ={
                    "RM" : 8,
                    "RV" : self.RHO,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF,
                }

                if self.UNINT_AD != 0:
                    bs_json.update({"W_TYPE" : 1 , "W_ANGLE" : self.UNINT_AD})

                return bs_json
            
            def __str__(self):
                return str(self.__dict__)
            
        class European:
            
            def __init__(self,rel_class,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0,unint_ang_disp=0):
                '''
                    rel_class =  1 Ordinary | 2 Low | 3 HotRolled
                '''
                self.CODE = 'European'
                self.CLASS = rel_class

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac

                self.UNINT_AD = unint_ang_disp
            
            def json(self):
                bs_json ={
                    "RM" : 5,
                    "RV" : self.CLASS,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF,
                }

                if self.UNINT_AD != 0:
                    bs_json.update({"W_TYPE" : 1 , "W_ANGLE" : self.UNINT_AD})

                return bs_json
            
            def __str__(self):
                return str(self.__dict__)

        class IRC_18:
            
            def __init__(self,factor,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0):

                self.CODE = 'IRC:18-2000'

                self.FACTOR = factor

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac
            
            def json(self):
                bs_json ={
                    "RM" : 4,
                    "RV" : self.FACTOR,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF
                }
                return bs_json
            
            def __str__(self):
                return str(self.__dict__)
            
        class IRC_112:
            
            def __init__(self,factor,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0):

                self.CODE = 'IRC:112-2011'


                self.FACTOR = factor

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac
            
            def json(self):
                bs_json ={
                    "RM" : 7,
                    "RV" : self.FACTOR,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF
                }
                return bs_json
            
            def __str__(self):
                return str(self.__dict__)
            
        class Null:
            
            def __init__(self,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0):

                self.CODE = 'No Relaxation'

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac
            
            def json(self):
                bs_json ={
                    "RM" : 0,
                    "RV" : 0,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF,
                    "bRELAX" : False
                }
                return bs_json
            
            def __str__(self):
                return str(self.__dict__)

        class Magura:
            
            def __init__(self,factor,ult_st,yield_st,curv_fric_fac=0,wob_fric_fac=0):

                self.CODE = 'Magura'

                if factor not in [10,45]:
                    factor = 45
                
                self.FACTOR = factor

                self.ULT_ST = ult_st
                self.YIELD_ST = yield_st
                self.CURV_FF = curv_fric_fac
                self.WOBBLE_FF = wob_fric_fac
            
            def json(self):
                bs_json ={
                    "RM" : 0,
                    "RV" : self.FACTOR,
                    "US" : self.ULT_ST,
                    "YS" : self.YIELD_ST,
                    "FF" : self.CURV_FF,
                    "WF" : self.WOBBLE_FF
                }
                return bs_json
            
            def __str__(self):
                return str(self.__dict__)
                                     
    # -----------------   TENDON    PROPERTY  --------------------------
    class Property:
        properties =[]
        ids = []

        def __init__(self,name,type,matID:int,tdn_area,duct_dia,relaxation,ext_mom_mag=0,anch_slip_begin=0,anch_slip_end=0,bond_type:bool=True,id=None):
            
            '''
            type = ['Internal (Pre-tension)' , 'Internal (Post-tenstion)' , 'External'] =>  1,2,3
            '''
            if Tendon.Property.ids == []: 
                tp_count = 1
            else:
                tp_count = max(Tendon.Property.ids)+1
            
            if id == None : self.ID = tp_count
            if id != None : self.ID = id

            self.NAME = name

            if type == 2:
                self.TYPE = 'INTERNAL'
                self.TENS = 'POST'
            elif type == 3:
                self.TYPE = 'EXTERNAL'
                self.TENS = 'PRE'
            else :
                self.TYPE = 'INTERNAL'
                self.TENS = 'PRE'

            self.MAT = matID
            self.TDN_AREA = tdn_area
            self.DUCT_DIA = duct_dia

            self.RELAX = relaxation

            self.EXT_MOM_MAG = ext_mom_mag
            self.ANC_SLIP_B = anch_slip_begin
            self.ANC_SLIP_E = anch_slip_end
            self.BOND_TYPE = bond_type

            Tendon.Property.properties.append(self)
            Tendon.Property.ids.append(self.ID)

        @classmethod
        def json(cls):
            json = {"Assign":{}}

            for self in cls.properties:
                json['Assign'][self.ID]={
                    "NAME" : self.NAME,
                    "TYPE" : self.TYPE,
                    "LT" : self.TENS,
                    "MATL" : self.MAT,
                    "AREA" : self.TDN_AREA,
                    "D_AREA" : self.DUCT_DIA,
                    "ASB" : self.ANC_SLIP_B,
                    "ASE" : self.ANC_SLIP_E,
                    "bBONDED" : self.BOND_TYPE,
                    "ALPHA" : self.EXT_MOM_MAG
                }
                json['Assign'][self.ID].update(self.RELAX.json())
            
            return json
        
        @classmethod
        def create(cls):
            MidasAPI("PUT","/db/TDNT",cls.json())


        @classmethod
        def get(cls):
            return MidasAPI('GET','/db/TDNT')
        
        @classmethod
        def delete(cls):
            MidasAPI("DELETE","/db/TDNT")
            cls.clear()
        
        @classmethod
        def clear(cls):
            cls.properties = []
            cls.ids = []


        @classmethod
        def sync(cls):
            a = cls.get()
            if a != {'message': ''}:
                if list(a['TDNT'].keys()) != []:
                    cls.properties = []
                    cls.ids=[]
                    for id in a['TDNT'].keys():
                        _JStoObj_Prop(int(id),a['TDNT'][id])


    # -----------------   TENDON    PROFILE  --------------------------
    class Profile:
        profiles =[]
        ids=[]

        def __init__(self,name,tdn_prop,tdn_group=0,elem=[],inp_type='3D',curve_type = 'SPLINE',st_len_begin = 0 , st_len_end = 0,n_typical_tendon=0,
                     trans_len_opt='USER', trans_len_begin = 0 , trans_len_end = 0, debon_len_begin=0 , debon_len_end=0,
                     ref_axis = 'ELEMENT',
                     prof_xyz = [], prof_xy =[],prof_xz=[],
                     prof_ins_point_end = 'END-I', prof_ins_point_elem = 0, x_axis_dir_element = 'I-J', x_axis_rot_ang = 0 , projection = True, offset_y = 0 , offset_z = 0,
                     prof_ins_point =[0,0,0], x_axis_dir_straight = 'X' , x_axis_dir_vec = [0,0], grad_rot_axis = 'X', grad_rot_ang=0,
                     radius_cen = [0,0], offset = 0, dir = 'CW',
                     id=None):
            '''
                TDN GROUP = Group ID
            '''



            if Tendon.Profile.ids == []: 
                td_count = 1
            else:
                td_count = max(Tendon.Profile.ids)+1
            
            if id == None : self.ID = td_count
            else : self.ID = id

            self.NAME = name
            self.PROP = tdn_prop
            self.GROUP = tdn_group
            self.ELEM = elem

            if inp_type not in ['2D' , '3D']: inp_type = '3D'
            self.INPUT = inp_type

            if curve_type not in ['SPLINE' , 'ROUND']: curve_type = 'ROUND'
            self.CURVE = curve_type

            self.BELENG = st_len_begin
            self.ELENG = st_len_end

            
            self.CNT = n_typical_tendon 
            if n_typical_tendon > 0: 
                self.bTP = True
            else: self.bTP = False

            if trans_len_opt not in ['USER' , 'AUTO']: trans_len_opt = 'USER'
            self.LENG_OPT = trans_len_opt
            self.BLEN = trans_len_begin
            self.ELEN =  trans_len_end

            self.DeBondBLEN = debon_len_begin
            self.DeBondELEN = debon_len_end

            if ref_axis not in ['ELEMENT' , 'STRAIGHT' , 'CURVE']: ref_axis = 'ELEMENT'
            self.SHAPE = ref_axis

            #------- ELEMENT TYPE -------------

            if prof_ins_point_end not in ['END-I' , 'END-J']: prof_ins_point_end = 'END-I'
            self.INS_PT = prof_ins_point_end

            if prof_ins_point_elem == 0: prof_ins_point_elem = elem[0]
            self.INS_ELEM = prof_ins_point_elem

            if x_axis_dir_element not in ['I-J' , 'J-I']: x_axis_dir_element = 'I-J'
            self.AXIS_IJ = x_axis_dir_element

            self.XAR_ANGLE = x_axis_rot_ang  # common in straight
            self.bPJ = projection # common in straight

            self.OFF_YZ = [offset_y,offset_z]

            #------- STRAIGHT TYPE -------------

            self.IP = prof_ins_point

            if x_axis_dir_straight not in ['X' , 'Y' , 'VECTOR']: x_axis_dir_straight = 'X'
            self.AXIS = x_axis_dir_straight

            self.VEC = x_axis_dir_vec


            if grad_rot_axis not in ['X' , 'Y']: grad_rot_axis = 'X'
            self.GR_AXIS = grad_rot_axis

            self.GR_ANGLE = grad_rot_ang

            #------- CURVE TYPE -------------

            self.RC = radius_cen
            self.OFFSET =  offset

            if dir not in ['CW' , 'CCW']: dir = 'CW'
            self.DIR = dir



            #---------------   PROFILES CREATION -----------------

            #----- 3D Profile (Round + Spline) -------------
            xyz_loc = []
            bFix = []
            R_spline3d = []
            R_round3d = []

            for point in prof_xyz:
                xyz_loc.append(_POINT_(point[0],point[1],point[2]))
                bFix.append(False) # Default not defining here
                if curve_type == 'SPLINE':
                    R_spline3d.append([0,0])   # Default not defining here
                else:
                    R_round3d.append(0)

            self.P_XYZ = xyz_loc

            self.bFIX = bFix
            self.R = R_spline3d
            self.RADIUS = R_round3d

            

            #----- 2D Profile Spline (only)-------------
            xy_loc = []
            xz_loc = []

            bFix_y = []
            bFix_z = []

            R_spline2d_Rz = []
            R_spline2d_Ry = []

            for point in prof_xy:
                xy_loc.append(_POINT_(point[0],point[1],0))
                bFix_y.append(False) # Default not defining here
                R_spline2d_Rz.append(0)

            for point in prof_xz:
                xz_loc.append(_POINT_(point[0],0,point[1]))
                bFix_z.append(False) # Default not defining here
                R_spline2d_Ry.append(0)


            self.P_XY = xy_loc
            self.P_XZ = xz_loc

            self.bFIX_XY = bFix_y
            self.bFIX_XZ = bFix_z

            self.R_XY = R_spline2d_Rz
            self.R_XZ = R_spline2d_Ry


            Tendon.Profile.profiles.append(self)
            Tendon.Profile.ids.append(self.ID)

        def update_profile(self,points_xyz):
            xyz_loc = []
            bFix = []
            R_spline3d = []
            R_round3d = []

            for point in points_xyz:
                xyz_loc.append(_POINT_(point[0],point[1],point[2]))
                bFix.append(False) # Default not defining here
                R_spline3d.append([0,0])   # Default not defining here

            self.P_XYZ = xyz_loc
            self.INPUT = '3D'
            self.CURVE = 'SPLINE'
            self.SHAPE = 'STRAIGHT'

            self.bFIX = bFix
            self.R = R_spline3d
            self.RADIUS = R_round3d

        @classmethod
        def json(cls):

            json = {"Assign":{}}

            for self in cls.profiles:
                json["Assign"][self.ID] = {
                                    'NAME' : self.NAME,
                                    'TDN_PROP' : self.PROP,
                                    'ELEM' : self.ELEM,
                                    'BELENG' : self.BELENG,
                                    'ELENG' : self.ELENG,
                                    'CURVE' : self.CURVE,
                                    'INPUT' : self.INPUT,
                                    'TDN_GRUP' : self.GROUP,
                                    "LENG_OPT": self.LENG_OPT,
                                    "BLEN": self.BLEN,
                                    "ELEN": self.ELEN,
                                    "bTP": self.bTP,
                                    "CNT": self.CNT,
                                    "DeBondBLEN": self.DeBondBLEN,
                                    "DeBondELEN": self.DeBondELEN,
                                    "SHAPE": self.SHAPE
                                }
                # --------------------------------   2D OR 3D (ROUND/SPLINE)--------------------------
                if self.INPUT == '3D':

                    # -------- 3D  ------------
                    array_temp = []

                    # -------- 3D SPLINE & ROUND ------------
                    if self.CURVE == 'ROUND' :
                        for j in range(len(self.P_XYZ)):
                            array_temp.append({
                                    'PT' : [self.P_XYZ[j].X,self.P_XYZ[j].Y,self.P_XYZ[j].Z],
                                    'bFIX' : self.bFIX[j],
                                    'RADIUS' : self.RADIUS[j]
                            })
                    else:
                        for j in range(len(self.P_XYZ)):
                            array_temp.append({
                                    'PT' : [self.P_XYZ[j].X,self.P_XYZ[j].Y,self.P_XYZ[j].Z],
                                    'bFIX' : self.bFIX[j],
                                    'R' : self.R[j]
                            })
                    
                    
                    
                    
                    # --- 3D Main ----

                    json_prof = {
                                    "PROF":array_temp
                                }

                elif self.INPUT == '2D':

                    # -------- 2D  ------------
                    array_y_temp = []
                    array_z_temp = []

                    # -------- 2D ONLY SPLINE  NOT ROUND ❌ ------------

                    for j in range(len(self.P_XY)):
                            array_y_temp.append({
                                    'PT' : [self.P_XY[j].X,self.P_XY[j].Y],
                                    'bFIX' : self.bFIX_XY[j],
                                    'R' : self.R_XY[j]
                            })

                    for j in range(len(self.P_XZ)):
                            array_z_temp.append({
                                    'PT' : [self.P_XZ[j].X,self.P_XZ[j].Z],
                                    'bFIX' : self.bFIX_XZ[j],
                                    'R' : self.R_XZ[j]
                            })

                    
                    # --- 3D Main ----

                    json_prof = {
                                    "PROFY":array_y_temp,
                                    "PROFZ":array_z_temp
                                }
                    
                # -------------------------------------------     TYPE  (ELEMNENT , STRAIGHT , CURVE)   ------------------------------------

                # ----- 3D Spline Element--------
                if self.SHAPE == 'ELEMENT' :
                    json_shape={
                                            "INS_PT": self.INS_PT,
                                            "INS_ELEM": self.INS_ELEM,
                                            "AXIS_IJ": self.AXIS_IJ,
                                            "XAR_ANGLE": self.XAR_ANGLE,
                                            "bPJ": self.bPJ,
                                            "OFF_YZ": self.OFF_YZ,
                                            }
                    
                # ----- 3D Spline Straight --------
                elif self.SHAPE == 'STRAIGHT' :
                    json_shape={
                                            "IP" : self.IP,
                                            "AXIS" : self.AXIS,
                                            "VEC" : self.VEC,
                                            "XAR_ANGLE": self.XAR_ANGLE,
                                            "bPJ": self.bPJ,
                                            "GR_AXIS": self.GR_AXIS,
                                            "GR_ANGLE": self.GR_ANGLE,
                                            }
                    
                # ----- 3D Spline Curve --------
                elif self.SHAPE == 'CURVE' :
                    json_shape={
                                            "IP" : self.IP,
                                            "RC" : self.RC,
                                            "OFFSET" : self.OFFSET,
                                            "DIR" : self.DIR,
                                            "XAR_ANGLE": self.XAR_ANGLE,
                                            "bPJ": self.bPJ,
                                            "GR_AXIS": self.GR_AXIS,
                                            "GR_ANGLE": self.GR_ANGLE,
                                            }
                
                json["Assign"][self.ID].update(json_shape)
                json["Assign"][self.ID].update(json_prof)
                        
            return json
        

        @classmethod
        def create(cls):
            MidasAPI("PUT","/db/TDNA",cls.json())


        @classmethod
        def get(cls):
            return MidasAPI('GET','/db/TDNA')
        
        @classmethod
        def delete(cls):
            cls.clear()
            return MidasAPI('DELETE','/db/TDNA')
        
        @classmethod
        def clear(cls):
            cls.profiles=[]
            cls.ids=[]
        
        @classmethod
        def sync(cls):
            a = cls.get()
            if a != {'message': ''}:
                if list(a['TDNA'].keys()) != []:
                    cls.profiles = []
                    cls.ids=[]
                    for id in a['TDNA'].keys():
                        _JStoObj_Profile(int(id),a['TDNA'][id])

    # ---------------------    END   CLASSS   -----------------------------------------------------


    class Prestress:
        """Prestress Loading for Tendons.
        """
        loads = []
        ids = []
        def __init__(self, profile_name, load_case, load_group = "", prestress_type = "STRESS", jack_step = "BEGIN", jack_begin = 0, jack_end=0, grouting_stage = 0, id = None):

            if id == None: id = 0
            if id > -1 :
                chk = 0
                for i in Load_Case.cases:
                    if load_case in i.NAME: chk = 1
                if chk == 0: Load_Case("PS", load_case)
                if load_group != "":
                    chk = 0
                    a = [v['NAME'] for v in Group.Load.json()["Assign"].values()]
                    if load_group in a: chk = 1
                    if chk == 0: Group.Load(load_group)


                if prestress_type not in ['STRESS' , 'FORCE']: prestress_type = 'STRESS'
                if jack_step not in ['BEGIN' , 'END' , 'BOTH']: jack_step = 'BEGIN'

                self.TDN_ID = 0
                if id == 0: id = len(Tendon.Prestress.loads) + 1
                self.ID = id
            else:
                self.TDN_ID = -id//100000
                self.ID = -id%100000
            

            self.TDN_NAME = profile_name

            self.LCNAME = load_case
            self.LDGR = load_group
            self.TYPE =prestress_type
            self.ORDER = jack_step

            self.JACK_BEGIN = jack_begin
            self.JACK_END = jack_end
            self.GROUTING = grouting_stage

            

            Tendon.Prestress.loads.append(self)
            Tendon.Prestress.ids.append(self.ID)
            

        @classmethod
        def json(cls):
            json = {"Assign": {}}
            for self in cls.loads:

                # Finding Tendon ID
                tdn_id = self.TDN_ID
                if not self.TDN_ID :
                    tdn_check=0
                    for prof in Tendon.Profile.profiles:
                        if prof.NAME == self.TDN_NAME:
                            tdn_id = prof.ID
                            tdn_check=1
                            break
                    if not tdn_check: 
                        print(f'⚠️   "{self.TDN_NAME}" Tendon name is not found for Prestress load application. Prestress load skipped.\n📑  Try Tendon.Profile.sync() to retrieve Profile Names first')
                        continue


                if tdn_id not in list(json["Assign"].keys()):
                    json["Assign"][tdn_id] = {"ITEMS": []}

                json["Assign"][tdn_id]["ITEMS"].append({
                                        "ID": self.ID,
                                        "LCNAME": self.LCNAME,
                                        "GROUP_NAME": self.LDGR,
                                        "TENDON_NAME": self.TDN_NAME,
                                        "TYPE": self.TYPE,
                                        "ORDER": self.ORDER,
                                        "BEGIN": self.JACK_BEGIN,
                                        "END": self.JACK_END,
                                        "GROUTING": self.GROUTING
                                    })
            return json
        
        @classmethod
        def create(cls):
            MidasAPI("PUT", "/db/TDPL",cls.json())

        @classmethod
        def get(cls):
            return MidasAPI("GET", "/db/TDPL")
        
        @classmethod
        def delete(cls):
            cls.clear()
            return MidasAPI("DELETE", "/db/TDPL")
        
        @classmethod
        def clear(cls):
            cls.loads=[]
        
        @classmethod
        def sync(cls):
            cls.loads = []
            a = cls.get()
            if a != {'message': ''}:
                for i in a['TDPL'].keys():
                    for j in range(len(a['TDPL'][i]['ITEMS'])):
                        Tendon.Prestress(a['TDPL'][i]['ITEMS'][j]['TENDON_NAME'],a['TDPL'][i]['ITEMS'][j]['LCNAME']
                                         ,a['TDPL'][i]['ITEMS'][j]['GROUP_NAME']
                                         ,a['TDPL'][i]['ITEMS'][j]['TYPE']
                                         ,a['TDPL'][i]['ITEMS'][j]['ORDER']
                                         ,a['TDPL'][i]['ITEMS'][j]['BEGIN']
                                         ,a['TDPL'][i]['ITEMS'][j]['END']
                                         ,a['TDPL'][i]['ITEMS'][j]['GROUTING']
                                         ,-(100000*int(i)+int(a['TDPL'][i]['ITEMS'][j]['ID'])))


from ._pscSS import _SS_PSC_12CELL,_SS_PSC_I,_SS_PSC_Value
from ._dbSecSS import _SS_DBUSER
from ._offsetSS import Offset
from ._unSupp import _SS_UNSUPP,_SS_STD_DB
from ._compositeSS import _SS_COMP_PSC_I,_SS_COMP_STEEL_I_TYPE1
from ._TapdbSecSS import _SS_TAPERED_DBUSER

from ._tapPSC12CellSS import _SS_TAP_PSC_12CELL

from midas_civil import MidasAPI


class _helperSECTION:
    ID, NAME, SHAPE, TYPE, OFFSET, USESHEAR, USE7DOF = 0,0,0,0,0,0,0
    def update():
        pass
    def toJSON():
        pass


def _SectionADD(self):
    # Commom HERE ---------------------------------------------
    
    id = int(self.ID)
    if id==None: id = 0
    
    if Section.ids == []: 
        count = 1
    else:
        count = max(Section.ids)+1

    if id==0 :
        self.ID = count
        Section.sect.append(self)
        Section.ids.append(int(self.ID))
    elif id in Section.ids:
        self.ID=int(id)
        print(f'⚠️  Section with ID {id} already exist! It will be replaced.')
        index=Section.ids.index(id)
        Section.sect[index]=self
    else:
        self.ID=id        
        Section.sect.append(self)
        Section.ids.append(int(self.ID))
    # Common END -------------------------------------------------------



def off_JS2Obj(js):

    try: OffsetPoint = js['OFFSET_PT']
    except: OffsetPoint='CC'
    try: CenterLocation = js['OFFSET_CENTER']
    except: CenterLocation=0
    try: HOffset = js['USERDEF_OFFSET_YI']
    except: HOffset=0
    try: HOffOpt = js['HORZ_OFFSET_OPT']
    except: HOffOpt=0
    try: VOffOpt = js['VERT_OFFSET_OPT']
    except: VOffOpt=0
    try: VOffset = js['USERDEF_OFFSET_ZI']
    except: VOffset=0
    try: UsrOffOpt = js['USER_OFFSET_REF']
    except: UsrOffOpt=0

    return Offset(OffsetPoint,CenterLocation,HOffset,HOffOpt,VOffset,VOffOpt,UsrOffOpt)



# -------------------  FUNCTION TO CREATE OBJECT used in ELEMENT SYNC  --------------------------
def _JS2OBJ(id,js):
    name = js['SECT_NAME']
    type = js['SECTTYPE']
    shape = js['SECT_BEFORE']['SHAPE']
    offset = off_JS2Obj(js['SECT_BEFORE'])
    uShear = js['SECT_BEFORE']['USE_SHEAR_DEFORM']
    u7DOF = js['SECT_BEFORE']['USE_WARPING_EFFECT']
    if type == 'DBUSER':
        if js['SECT_BEFORE']['DATATYPE'] ==2: obj = _SS_DBUSER._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        else: obj = _SS_STD_DB(id,name,type,shape,offset,uShear,u7DOF,js)

    elif type == 'PSC' :
        if shape in ['1CEL','2CEL']: obj = _SS_PSC_12CELL._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        elif shape in ['PSCI']: obj = _SS_PSC_I._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        elif shape in ['VALU']: obj = _SS_PSC_Value._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        else: obj = _SS_UNSUPP(id,name,type,shape,offset,uShear,u7DOF,js)

    elif type == 'COMPOSITE':
        if shape in ['CI']: obj = _SS_COMP_PSC_I._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        elif shape in ['I']: obj = _SS_COMP_STEEL_I_TYPE1._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        else: obj = _SS_UNSUPP(id,name,type,shape,offset,uShear,u7DOF,js)

    elif type == 'TAPERED' :
        try:
            typeDB = js['SECT_BEFORE']['TYPE']
        except: typeDB = 0
        if typeDB == 2: 
            obj = _SS_TAPERED_DBUSER._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        elif shape in ['1CEL','2CEL']: obj = _SS_TAP_PSC_12CELL._objectify(id,name,type,shape,offset,uShear,u7DOF,js)
        else: obj = _SS_UNSUPP(id,name,type,shape,offset,uShear,u7DOF,js)

    else :
        obj = _SS_UNSUPP(id,name,type,shape,offset,uShear,u7DOF,js)


    _SectionADD(obj)





class Section:
    """ NEW Create Sections \n Use Section.USER , Section.PSC to create sections"""
    sect:list[_helperSECTION] = []
    ids:list[int] = []


    @classmethod
    def json(cls):
        json = {"Assign":{}}
        for sect in cls.sect:
            js = sect.toJSON()
            json["Assign"][str(sect.ID)] = js
        return json
    
    @staticmethod
    def create():
        MidasAPI("PUT","/db/SECT",Section.json())


    @staticmethod
    def get():
        return MidasAPI("GET","/db/SECT")
    
    
    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/SECT")
        Section.clear()

    @staticmethod
    def clear():
        Section.sect=[]
        Section.ids=[]


    @staticmethod
    def sync():
        a = Section.get()
        if a != {'message': ''}:
            Section.sect = []
            Section.ids=[]
            for sect_id in a['SECT'].keys():
                _JS2OBJ(sect_id,a['SECT'][sect_id])


#---------------------------------     S E C T I O N S    ---------------------------------------------

    #---------------------     D B   U S E R    --------------------
    @staticmethod
    def DBUSER(Name='',Shape='',parameters:list=[],Offset=Offset(),useShear=True,use7Dof=False,id:int=None): 
        args = locals()
        sect_Obj = _SS_DBUSER(**args)
        _SectionADD(sect_Obj)
        return sect_Obj
    
    class PSC :

        @staticmethod
        def CEL12(Name='',Shape='1CEL',Joint=[0,0,0,0,0,0,0,0],
                    HO1=0,HO2=0,HO21=0,HO22=0,HO3=0,HO31=0,
                    BO1=0,BO11=0,BO12=0,BO2=0,BO21=0,BO3=0,
                    HI1=0,HI2=0,HI21=0,HI22=0,HI3=0,HI31=0,HI4=0,HI41=0,HI42=0,HI5=0,
                    BI1=0,BI11=0,BI12=0,BI21=0,BI3=0,BI31=0,BI32=0,BI4=0,
                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=None):
            args = locals()
            sect_Obj = _SS_PSC_12CELL(**args)
            _SectionADD(sect_Obj)
            return sect_Obj
        
        @staticmethod
        def I(Name='',Symm = True,Joint=[0,0,0,0,0,0,0,0,0],
                            H1=0,
                            HL1=0,HL2=0,HL21=0,HL22=0,HL3=0,HL4=0,HL41=0,HL42=0,HL5=0,
                            BL1=0,BL2=0,BL21=0,BL22=0,BL4=0,BL41=0,BL42=0,
                            HR1=0,HR2=0,HR21=0,HR22=0,HR3=0,HR4=0,HR41=0,HR42=0,HR5=0,
                            BR1=0,BR2=0,BR21=0,BR22=0,BR4=0,BR41=0,BR42=0,
                            Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=None):
             
            args = locals()
            sect_Obj = _SS_PSC_I(**args)
            
            _SectionADD(sect_Obj)
            return sect_Obj
        
        @staticmethod
        def Value(Name:str,
                    OuterPolygon:list,InnerPolygon:list=[],
                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=None):
             
            args = locals()
            sect_Obj = _SS_PSC_Value(**args)
            
            _SectionADD(sect_Obj)
            return sect_Obj
        
    
    class Composite :
        @staticmethod
        def PSCI(Name='',Symm = True,Joint=[0,0,0,0,0,0,0,0,0],
                    Bc=0,tc=0,Hh=0,
                    H1=0,
                    HL1=0,HL2=0,HL21=0,HL22=0,HL3=0,HL4=0,HL41=0,HL42=0,HL5=0,
                    BL1=0,BL2=0,BL21=0,BL22=0,BL4=0,BL41=0,BL42=0,
                    HR1=0,HR2=0,HR21=0,HR22=0,HR3=0,HR4=0,HR41=0,HR42=0,HR5=0,
                    BR1=0,BR2=0,BR21=0,BR22=0,BR4=0,BR41=0,BR42=0,
                    EgdEsb =0, DgdDsb=0,Pgd=0,Psb=0,TgdTsb=0,
                    MultiModulus = False,CreepEratio=0,ShrinkEratio=0,
                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=None):
             
            args = locals()
            sect_Obj = _SS_COMP_PSC_I(**args)
            
            _SectionADD(sect_Obj)
            return sect_Obj
        
        @staticmethod
        def SteelI_Type1(Name='',Bc=0,tc=0,Hh=0,Hw=0,B1=0,tf1=0,tw=0,B2=0,tf2=0,EsEc =0, DsDc=0,Ps=0,Pc=0,TsTc=0,
                MultiModulus = False,CreepEratio=0,ShrinkEratio=0,
                Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=None):
             
            args = locals()
            sect_Obj = _SS_COMP_STEEL_I_TYPE1(**args)
            
            _SectionADD(sect_Obj)
            return sect_Obj
    
    class Tapered:

        @staticmethod
        def DBUSER(Name='',Shape='',params_I:list=[],params_J:list=[],Offset=Offset(),useShear=True,use7Dof=False,id:int=None):
            args = locals()
            sect_Obj = _SS_TAPERED_DBUSER(**args)
            
            _SectionADD(sect_Obj)
            return sect_Obj
        
        @staticmethod
        def PSC12CEL(Name='',Shape='1CEL',Joint=[0,0,0,0,0,0,0,0],
                    HO1_I=0,HO2_I=0,HO21_I=0,HO22_I=0,HO3_I=0,HO31_I=0,
                    BO1_I=0,BO11_I=0,BO12_I=0,BO2_I=0,BO21_I=0,BO3_I=0,
                    HI1_I=0,HI2_I=0,HI21_I=0,HI22_I=0,HI3_I=0,HI31_I=0,HI4_I=0,HI41_I=0,HI42_I=0,HI5_I=0,
                    BI1_I=0,BI11_I=0,BI12_I=0,BI21_I=0,BI3_I=0,BI31_I=0,BI32_I=0,BI4_I=0,

                    HO1_J=0,HO2_J=0,HO21_J=0,HO22_J=0,HO3_J=0,HO31_J=0,
                    BO1_J=0,BO11_J=0,BO12_J=0,BO2_J=0,BO21_J=0,BO3_J=0,
                    HI1_J=0,HI2_J=0,HI21_J=0,HI22_J=0,HI3_J=0,HI31_J=0,HI4_J=0,HI41_J=0,HI42_J=0,HI5_J=0,
                    BI1_J=0,BI11_J=0,BI12_J=0,BI21_J=0,BI3_J=0,BI31_J=0,BI32_J=0,BI4_J=0,

                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=None):
            args = locals()
            sect_Obj = _SS_TAP_PSC_12CELL(**args)
            
            _SectionADD(sect_Obj)
            return sect_Obj

        
        

#---------------------------------     T A P E R E D   G R O U P    ---------------------------------------------
    class TaperedGroup:
        
        data = []
        
        def __init__(self, name, elem_list, z_var="LINEAR", y_var="LINEAR", z_exp=2.0, z_from="i", z_dist=0, 
                     y_exp=2.0, y_from="i", y_dist=0, id=""):
            """
            Args:
                name (str): Tapered Group Name (Required).
                elem_list (list): List of element numbers (Required).
                z_var (str): Section shape variation for Z-axis: "LINEAR" or "POLY" (Required).
                y_var (str): Section shape variation for Y-axis: "LINEAR" or "POLY" (Required).
                z_exp (float, optional): Z-axis exponent. Required if z_var is "POLY".
                z_from (str, optional): Z-axis symmetric plane ("i" or "j"). Defaults to "i" for "POLY".
                z_dist (float, optional): Z-axis symmetric plane distance. Defaults to 0 for "POLY".
                y_exp (float, optional): Y-axis exponent. Required if y_var is "POLY".
                y_from (str, optional): Y-axis symmetric plane ("i" or "j"). Defaults to "i" for "POLY".
                y_dist (float, optional): Y-axis symmetric plane distance. Defaults to 0 for "POLY".
                id (str, optional): ID for the tapered group. Auto-generated if not provided.
            
            Example:
                Section.TapperGroup("Linear", [1, 2, 3], "LINEAR", "LINEAR")
                Section.TapperGroup("ZPoly", [4, 5], "POLY", "LINEAR", z_exp=2.5)
            """
            self.NAME = name
            self.ELEM_LIST = elem_list
            self.Z_VAR = z_var
            self.Y_VAR = y_var
            
            # Z-axis parameters (only for POLY)
            if z_var == "POLY":
                if z_exp is None:
                    raise ValueError("z_exp is required when z_var is 'POLY'")
                self.Z_EXP = z_exp
                self.Z_FROM = z_from if z_from is not None else "i"
                self.Z_DIST = z_dist if z_dist is not None else 0
            else:
                self.Z_EXP = None
                self.Z_FROM = None
                self.Z_DIST = None
            
            # Y-axis parameters (only for POLY)
            if y_var == "POLY":
                if y_exp is None:
                    raise ValueError("y_exp is required when y_var is 'POLY'")
                self.Y_EXP = y_exp
                self.Y_FROM = y_from if y_from is not None else "i"
                self.Y_DIST = y_dist if y_dist is not None else 0
            else:
                self.Y_EXP = None
                self.Y_FROM = None
                self.Y_DIST = None
            
            if id == "":
                id = len(Section.TaperedGroup.data) + 1
            self.ID = id
            
            Section.TaperedGroup.data.append(self)
        
        @classmethod
        def json(cls):
            json_data = {"Assign": {}}
            for i in cls.data:
                # Base data that's always included
                tapper_data = {
                    "NAME": i.NAME,
                    "ELEMLIST": i.ELEM_LIST,
                    "ZVAR": i.Z_VAR,
                    "YVAR": i.Y_VAR
                }
                
                # Add Z-axis polynomial parameters only if Z_VAR is "POLY"
                if i.Z_VAR == "POLY":
                    tapper_data["ZEXP"] = i.Z_EXP
                    tapper_data["ZFROM"] = i.Z_FROM
                    tapper_data["ZDIST"] = i.Z_DIST
                
                # Add Y-axis polynomial parameters only if Y_VAR is "POLY"
                if i.Y_VAR == "POLY":
                    tapper_data["YEXP"] = i.Y_EXP
                    tapper_data["YFROM"] = i.Y_FROM
                    tapper_data["YDIST"] = i.Y_DIST
                
                json_data["Assign"][str(i.ID)] = tapper_data
            
            return json_data
        
        @classmethod
        def create(cls):
            MidasAPI("PUT", "/db/tsgr", cls.json())
        
        @classmethod
        def get(cls):
            return MidasAPI("GET", "/db/tsgr")
        
        @classmethod
        def delete(cls):
            cls.clear()
            return MidasAPI("DELETE", "/db/tsgr")
        
        @classmethod
        def clear(cls):
            cls.data = []
        
        @classmethod
        def sync(cls):
            cls.data = []
            response = cls.get()
            
            if response and response != {'message': ''}:
                tsgr_data = response.get('TSGR', {})
                # Iterate through the dictionary of tapered groups from the API response
                for tsgr_id, item_data in tsgr_data.items():
                    # Extract base parameters
                    name = item_data.get('NAME')
                    elem_list = item_data.get('ELEMLIST')
                    z_var = item_data.get('ZVAR')
                    y_var = item_data.get('YVAR')
                    
                    # Extract optional parameters based on variation type
                    z_exp = item_data.get('ZEXP') if z_var == "POLY" else None
                    z_from = item_data.get('ZFROM') if z_var == "POLY" else None
                    z_dist = item_data.get('ZDIST') if z_var == "POLY" else None
                    
                    y_exp = item_data.get('YEXP') if y_var == "POLY" else None
                    y_from = item_data.get('YFROM') if y_var == "POLY" else None
                    y_dist = item_data.get('YDIST') if y_var == "POLY" else None
                    
                    Section.TaperedGroup(
                        name=name,
                        elem_list=elem_list,
                        z_var=z_var,
                        y_var=y_var,
                        z_exp=z_exp,
                        z_from=z_from,
                        z_dist=z_dist,
                        y_exp=y_exp,
                        y_from=y_from,
                        y_dist=y_dist,
                        id=tsgr_id
                    )


from ._mapi import MidasAPI
# from ._model import *


#28 Class to generate load combinations
class LoadCombination:
    data = []
    valid = ["General", "Steel", "Concrete", "SRC", "Composite Steel Girder", "Seismic", "All"]
    com_map = {
            "General": "/db/LCOM-GEN",
            "Steel": "/db/LCOM-STEEL",
            "Concrete": "/db/LCOM-CONC",
            "SRC": "/db/LCOM-SRC",
            "Composite Steel Girder": "/db/LCOM-STLCOMP",
            "Seismic": "/db/LCOM-SEISMIC"
        }
    def __init__(self, name, case, classification = "General", active = "ACTIVE", typ = "Add", id:int = None, desc = ""):
        """Name, List of tuple of load cases & factors, classification, active, type. \n
        Sample: LoadCombination('LCB1', [('Dead Load(CS)',1.5), ('Temperature(ST)',0.9)], 'General', 'Active', 'Add')"""
        if id == None: id =0
        if not isinstance(case, list):
            print("case should be a list that contains tuple of load cases & factors.\nEg: [('Load1(ST)', 1.5), ('Load2(ST)',0.9)]")
            return
        for i in case:
            if not isinstance(i, tuple):
                print(f"{i} is not a tuple.  case should be a list that contains tuple of load cases & factors.\nEg: [('Load1(ST)', 1.5), ('Load2(ST)',0.9)]")
                return
            if not isinstance(i[0], str):
                print(f"{i[0]} is not a string.  case should be a list that contains tuple of load cases & factors.\nEg: [('Load1(ST)', 1.5), ('Load2(ST)',0.9)]")
                return
            if i[0][-1] != ")":
                print(f"Load case type is not mentioned for {i[0]}.  case should be a list that contains tuple of load cases & factors.\nEg: [('Load1(ST)', 1.5), ('Load2(ST)',0.9)]")
                return
            if not isinstance(i[1],(int, float)):
                print(f"{i[1]} is not a number.  case should be a list that contains tuple of load cases & factors.\nEg: [('Load1(ST)', 1.5), ('Load2(ST)',0.9)]")
                return

        if classification not in LoadCombination.valid[:-1]:
            print(f'"{classification}" is not a valid input.  It is changed to "General".')
            classification = "General"
            
        if classification in ["General", "Seismic"]:
            if active not in ["ACTIVE", "INACTIVE"]: active = "ACTIVE"
        if classification in  ["Steel", "Concrete", "SRC", "Composite Steel Girder"]:
            if active not in ["STRENGTH", "SERVICE", "INACTIVE"]: active = "STRENGTH"
        
        typ_map = {"Add": 0, "Envelope": 1, "ABS": 2, "SRSS": 3, 0:0, 1:1, 2:2, 3:3}
        if typ not in list(typ_map.keys()): typ = "Add"
        if classification not in ["General", "Seismic"] and typ_map.get(typ) == 2: typ = "Add"
        
        if id == 0 and len(LoadCombination.data) == 0: 
            id = 1
        elif id == 0 and len(LoadCombination.data) != 0:
            id = max([i.ID for i in LoadCombination.data]) + 1
        elif id != 0 and id in [i.ID for i in LoadCombination.data]:
            if classification in [i.CLS for i in LoadCombination.data if i.ID == id]:
                print(f"ID {id} is already defined.  Existing combination would be replaced.")
                
        
        combo = []
        valid_anl = ["ST", "CS", "MV", "SM", "RS", "TH", "CB", "CBC", "CBS", "CBR", "CBSC", "CBSM"] #Need to figure out for all combination types
        for i in case:
            a = i[0].rsplit('(', 1)[1].rstrip(')')
            if a in valid_anl:
                combo.append({
                    "ANAL": a,
                    "LCNAME":i[0].rsplit('(', 1)[0],
                    "FACTOR": i[1]
                })
        self.NAME = name
        self.CASE = combo
        self.CLS = classification
        self.ACT = active
        self.TYPE = typ_map.get(typ)
        self.ID = id
        self.DESC = desc
        LoadCombination.data.append(self)
    
    @classmethod
    def json(cls, classification = "All"):
        if len(LoadCombination.data) == 0:
            print("No Load Combinations defined!  Define the load combination using the 'LoadCombination' class before making json file.")
            return
        if classification not in LoadCombination.valid:
            print(f'"{classification}" is not a valid input.  It is changed to "General".')
            classification = "General"
        json = {k:{'Assign':{}} for k in LoadCombination.valid[:-1]}
        for i in LoadCombination.data:
            if i.CLS == classification or classification == "All":
                json[i.CLS]['Assign'][i.ID] = {
                    "NAME": i.NAME,
                    "ACTIVE": i.ACT,
                    "iTYPE": i.TYPE,
                    "DESC": i.DESC,
                    "vCOMB":i.CASE
                }
        json = {k:v for k,v in json.items() if v != {'Assign':{}}}
        return json
    
    @classmethod
    def get(cls, classification = "All"):
        if classification not in LoadCombination.valid:
            print(f'"{classification}" is not a valid input.  It is changed to "General".')
            classification = "General"
        combos = {k:{} for k in LoadCombination.valid[:-1]}
        for i in LoadCombination.valid[:-1]:
            if classification == i or classification == "All":
                combos[i] = MidasAPI("GET",LoadCombination.com_map.get(i))
        json = {k:v for k,v in combos.items() if v != {'message':''}}
        return json
    
    @classmethod
    def create(cls, classification = "All"):
        if len(LoadCombination.data) == 0:
            # print("No Load Combinations defined!  Define the load combination using the 'LoadCombination' class before creating these in the model.")
            return
        if classification not in LoadCombination.valid:
            print(f'"{classification}" is not a valid input.  It is changed to "General".')
            classification = "General"
        json = LoadCombination.json(classification)
        for i in LoadCombination.valid[:-1]:
            if classification == i or classification == "All":
                if i in list(json.keys()):
                    a = list(json[i]['Assign'].keys())
                    b=""
                    for j in range(len(a)):
                        b += str(a[j]) + ","
                    if b != "": b = "/" + b[:-1]
                    MidasAPI("DELETE", LoadCombination.com_map.get(i) + b)     #Delete existing combination if any
                    MidasAPI("PUT", LoadCombination.com_map.get(i), json[i])   #Create new combination
    
    @classmethod
    def sync(cls, classification = "All"):
        json = LoadCombination.get(classification)
        if json != {}:
            keys = list(json.keys())
            for i in keys:
                for k,v in json[i][LoadCombination.com_map.get(i)[4:]].items():
                    c = []
                    for j in range(len(v['vCOMB'])):
                        c.append((v['vCOMB'][j]['LCNAME'] + "("+ v['vCOMB'][j]['ANAL'] + ")", v['vCOMB'][j]['FACTOR']))
                    LoadCombination(v['NAME'], c, i, v['ACTIVE'], v['iTYPE'], int(k), v['DESC'])
    
    @classmethod
    def delete(cls, classification = "All", ids = []):
        json = LoadCombination.json(classification)
        a = ""
        for i in range(len(ids)):
            a += str(ids[i]) + ","
        a = "/" + a[:-1]
        if json == {}: 
            print("No load combinations are defined to delete.")
        for i in list(json.keys()):
            MidasAPI("DELETE",LoadCombination.com_map.get(i) + a)

    @classmethod
    def clear(cls):
        cls.data = []
#---------------------------------------------------------------------------------------------------------------


# from ._model import *
# from ._mapi import *
from __future__ import annotations
from math import hypot,sqrt
import numpy as np


#Function to remove duplicate set of values from 2 lists
# def unique_lists(li1, li2):
#     if type (li1) == list and type (li2) == list:
#         if len(li1) == len(li2):
#             indices_to_remove = []
#             for i in range(len(li1)):
#                 for j in range(i+1,len(li1)):
#                     if li1[i] == li1[j] and li2[i] == li2[j]:
#                         indices_to_remove.append(j)
#             for index in sorted(indices_to_remove, reverse = True):
#                 del li1[index]
#                 del li2[index]


# def sect_inp(sec):
#     """Section ID.  Enter one section id or list of section IDs.  Sample:  sect_inp(1) OR sect_inp([3,2,5])"""
#     Model.units()
#     a = MidasAPI("GET","/db/SECT",{"Assign":{}})
#     if type(sec)==int: sec = [sec]
#     b={}
#     for s in sec:
#         if str(s) in a['SECT'].keys() : b.update({s : a['SECT'][str(s)]})
#     # if elem = [0] and sec!=0: b.update({sec : })
#     if b == {}: b = "The required section ID is not defined in connected model file."
#     return(b)
#---------------------------------------------------------------------------------------------------------------



def sFlatten(list_of_list):
    # list_of_list = [list_of_list]
    return [item for elem in list_of_list for item in (elem if isinstance(elem, list) else [elem])]

# def getID_orig(element_list):
#     """Return ID of Node and Element"""
#     return [beam.ID for beam in sFlatten(element_list)]

def getID(*objects):
    objects = list(objects)
    _getID2(objects)
    return objects

def _getID2(objects):
    for i in range(len(objects)):
        if isinstance(objects[i], list):
            _getID2(objects[i])  # Recursive call for sublist
        else:
            objects[i] = objects[i].ID



def getNodeID(*objects):
    objects = list(objects)
    _getNodeID2(objects)
    return objects

def _getNodeID2(objects):
    for i in range(len(objects)):
        if isinstance(objects[i], list):
            _getNodeID2(objects[i])  # Recursive call for sublist
        else:
            objects[i] = objects[i].NODES




# def getNodeID_orig(element_list):
#     """Return Node IDs of Element"""
#     # return list(sFlatten([beam.NODES for beam in sFlatten(element_list)]))
#     return list(sFlatten([beam.NODES for beam in sFlatten(element_list)]))


def arr2csv(nlist):
    strinff = ",".join(map(str,nlist))
    return strinff

def zz_add_to_dict(dictionary, key, value):
    if key in dictionary:
        dictionary[key].append(value)
    else:
        dictionary[key] = [value]


def _convItem2List(item):
    if isinstance(item,(list,np.ndarray)):
        return item
    return [item]

def _matchArray(A,B):
    '''Matches B to length of A   
    Return B'''
    A = _convItem2List(A)
    B = _convItem2List(B)
    n = len(A)
    if len(B) >= n:
        return B[:n]
    return B + [B[-1]] * (n - len(B))

def _longestList(A,B):
    """ Matches A , B list and returns the list with longest length with last element repeated """
    A = _convItem2List(A)
    B = _convItem2List(B)
    nA = len(A)
    nB = len(B)

    if nA >= nB:
        return (A , B + [B[-1]] * (nA - nB))
    return (A + [A[-1]] * (nB - nA),B)




class utils:
    ''' Contains helper function and utilities'''
    class Alignment:
        '''Defines alignment object passing through the points'''
        
        def __init__(self,points,type: str = 'cubic'):
            ''' 
            **POINTS** -> Points on the alignment [[x,y] , [x,y] , [x,y] ....]   
            **TYPE** -> Type of interpolating curve
                    cubic , akima , makima , pchip
            '''
            from scipy.interpolate import CubicSpline , Akima1DInterpolator , PchipInterpolator

            _pt_x = [pt[0] for pt in points]
            _pt_y = [pt[1] for pt in points]

            # _alignment = splrep(_pt_x, _pt_y)
            if type == 'akima':
                _alignment = Akima1DInterpolator(_pt_x, _pt_y,method='akima')
            elif type == 'makima':
                _alignment = Akima1DInterpolator(_pt_x, _pt_y,method='makima')
            elif type == 'pchip':
                _alignment = PchipInterpolator(_pt_x, _pt_y)
            else :
                _alignment = CubicSpline(_pt_x, _pt_y)

            # _alignSlope = Akima1DInterpolator(_pt_x, _pt_y,method='akima') # Used for slope calculation

            _n=500
            # INITIAL ALGINMENT - Mapping U parameter to X (based on Distance)
            _x_fine = np.linspace(_pt_x[0],_pt_x[-1],_n)

            _y_fine = _alignment(_x_fine)

            _dx = np.diff(_x_fine)
            _dy = np.diff(_y_fine)

            _dl=[]
            for i in range(len(_dx)):
                _dl.append(hypot(_dx[i],_dy[i]))

            _cumLength = np.insert(np.cumsum(_dl),0,0)
            _totalLength = _cumLength[-1]

            _u_fine = _cumLength/_totalLength

            self.ALIGNMENT = _alignment
            # self.ALIGNSLOPE = _alignSlope
            self.TOTALLENGTH = _totalLength
            self.CUMLENGTH = _cumLength
            self.PT_X = _pt_x
            self.PT_Y = _pt_y
            self.X_FINE = _x_fine
            self.Y_FINE = _y_fine
            self.U_FINE = _u_fine

        @staticmethod
        def transformPoint(point:tuple,initial_align:utils.Alignment,final_align:utils.Alignment) -> list :
            ''' 
            Transforms a point (x,y) => [X , Y]    
            Maps a point (x,y) wrt Initial alignment curve to a new Final alignment (X,Y)
            '''
            ptx = point[0]
            pty = point[1]
            distx = 100000 #Initial high distance
            idx = 0
            y_ref = 0
            fact = 10000
            for q in range(101):
                x_onLine1 = ptx+initial_align.TOTALLENGTH*(q-50)/fact
                if x_onLine1 < initial_align.PT_X[0]:
                    continue
                if x_onLine1 > initial_align.PT_X[-1]:
                    break
                # y_onLine1 = splev(x_onLine1, initial_align.ALIGNMENT)
                y_onLine1 = initial_align.ALIGNMENT(x_onLine1)
                dist = hypot(ptx-x_onLine1,pty-y_onLine1)
                if dist <= distx:
                    distx = dist
                    idx = q
                    y_ref = y_onLine1
                # print(f"  > X location of line = {x_onLine1}  Y on Line = {y_onLine1}|   Distance = {dist}  |  Index = {q}")

            final_u = np.interp(ptx+initial_align.TOTALLENGTH*(idx-50)/fact,initial_align.X_FINE,initial_align.U_FINE)
            off = np.sign(pty-y_ref)*distx
            x2_interp = np.interp(final_u,final_align.U_FINE,final_align.X_FINE)

            # y2_interp = splev(x2_interp, final_align.ALIGNMENT)
            y2_interp = final_align.ALIGNMENT(x2_interp)

            slope = final_align.ALIGNMENT(x2_interp,1) # Tan theta

            norm = sqrt(1+slope*slope)
            x_off = -slope/norm
            y_off = 1/norm

            # print(f"Point loc = [{point}] , Index match = {idx} , Point X on Initial = {ptx+initial_align.TOTALLENGTH*(idx-50)/8000} , Point Y = {y_ref} , Distance = {off} , Xoff = {slope}")

            return (round(x2_interp+x_off*off,5),round(y2_interp+y_off*off,5))

    
    
    @staticmethod
    def LineToPlate(nDiv:int = 10 , mSizeDiv:float = 0, bRigdLnk:bool=True , meshSize:float=0.5, elemList:list=None):
        '''
        Converts selected/entered line element to Shell elements   
        **nDiv** - No. of Division along the length of span    
        **mSizeDiv** - Division based on mesh size(in meter) along the length of span   
                division based on number -> **mSizeDiv  = 0**  
                division based on meshSize(in meter) -> **nDiv = 0**   
        **bRigdLnk** - Whether to create Rigid links at the span ends  
        **meshSize** - Mesh size(in meter) of the plate elements   
        **elemList** - Element list which are to be converted . If None is passed, element are taken from selected elements in CIVIL NX  

        '''
        from ._utilsFunc._line2plate import SS_create
        SS_create(nDiv , mSizeDiv , bRigdLnk , meshSize ,elemList)



from setuptools import setup,find_packages

with open('README.md','r') as f:
    description = f.read()


setup(name='midas_civil',
    version="1.2.9",
    description='Python library for MIDAS Civil NX',
    author='Sumit Shekhar',
    author_email='sumit.midasit@gmail.com',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'polars',
        'xlsxwriter',
        'requests',
        'scipy',
        'colorama',
        'openpyxl',
        'tqdm'
    ],          
    long_description= description,
    long_description_content_type='text/markdown',
    url='https://github.com/MIDASIT-Co-Ltd/midas-civil-python',
    keywords=['midas','civil','civil nx','bridge'],
    license='MIT',
    )


from ._mapi import MidasAPI
#--------------------------------------------------------------------------------------------------

class AnalysisControl:
    
    class MainControlData:
        
        data = []
        
        def __init__(self, 
                    ardc: bool = True,
                    anrc: bool = True,
                    iter: int = 20,
                    tol: float = 0.001,
                    csecf: bool = False,
                    trs: bool = True,
                    crbar: bool = False,
                    bmstress: bool = False,
                    clats: bool = False):
            """
            Main Control Data constructor for analysis control settings.
            
            Parameters:
                ardc: Auto Rotational DOF Constraint for Truss/Plane Stress/Solid Elements (default True)
                
                anrc: Auto Normal Rotation Constraint for Plate Elements (default True)

                iter: Number of Iterations/Load Case (default 20)

                tol: Convergence Tolerance (default 0.001)
                
                csecf: Consider Section Stiffness Scale Factor for Stress Calculation (default False)

                trs: Transfer Reactions of Slave Node to the Master Node (default True)

                crbar: Consider Reinforcement for Section Stiffness Calculation (default False)

                bmstress: Calculate Equivalent Beam Stresses (Von-Mises and Max-Shear) (default False)
                
                clats: Change Local Axis of Tapered Section for Force/Stress Calculation (default False)
            
            Examples:
                # Basic control data with required parameters
                MainControlData(iter=20, tol=0.001)
                
                # with multiple options enabled
                MainControlData(
                    ardc=True, anrc=True, iter=30, tol=0.0005,
                    trs=True, bmstress=True
                )
            """
            
            # Validate required parameters
            if iter is None:
                raise ValueError("iter (Number of Iterations) is required")
            if tol is None:
                raise ValueError("tol (Convergence Tolerance) is required")
            
            # ID is always 1
            self.ID = 1
            
            # Set parameters
            self.ARDC = ardc
            self.ANRC = anrc
            self.ITER = iter
            self.TOL = tol
            self.CSECF = csecf
            self.TRS = trs
            self.CRBAR = crbar
            self.BMSTRESS = bmstress
            self.CLATS = clats
            
            # Add to static list
            AnalysisControl.MainControlData.data.append(self)
            
            # Automatically execute the data when instance is created
            self._execute()
        
        def _execute(self):
            """
            Automatically sends the MainControlData to the analysis system when created
            """
            json_data = {"Assign": {}}
            
            control_data = {
                "ARDC": self.ARDC,
                "ANRC": self.ANRC,
                "ITER": self.ITER,
                "TOL": self.TOL,
                "CSECF": self.CSECF,
                "TRS": self.TRS,
                "CRBAR": self.CRBAR,
                "BMSTRESS": self.BMSTRESS,
                "CLATS": self.CLATS
            }
            
            json_data["Assign"][str(self.ID)] = control_data
            
            MidasAPI("PUT", "/db/actl", json_data)

    class PDelta:
        """Create P-Delta Analysis Control Object in Python"""
        data = []
        
        def __init__(self, 
                    iter: int = 5,
                    tol: float = 0.00001,
                    load_case_data: list = None):
            """
            P-Delta Analysis Control constructor for geometric nonlinear analysis settings.
            
            Parameters:
                iter: Number of Iterations (default 5)
                
                tol: Convergence Tolerance (default 0.00001)

                load_case_data: Load Cases with Scale Factors (required)
                    - List of load cases and their corresponding scale factors for P-Delta analysis
                    - Format: [["LC1", factor1], ["LC2", factor2], ...]

            Example:
        
                PDelta(iter=5, load_case_data=[["DL", 1.0]])   
    
            """
            
            # Validate required parameters
            if iter is None:
                raise ValueError("iter (Number of Iterations) is required")
            if load_case_data is None or len(load_case_data) == 0:
                raise ValueError("load_case_data (Load Cases) is required")
            
            # Validate load case data format
            for i, case in enumerate(load_case_data):
                if not isinstance(case, list) or len(case) != 2:
                    raise ValueError(f"load_case_data[{i}] must be a list with 2 elements [name, factor]")
                if not isinstance(case[0], str):
                    raise ValueError(f"load_case_data[{i}][0] (load case name) must be a string")
                if not isinstance(case[1], (int, float)):
                    raise ValueError(f"load_case_data[{i}][1] (scale factor) must be a number")
            
            # ID is always 1
            self.ID = 1
            
            # Set parameters
            self.ITER = iter
            self.TOL = tol
            self.LOAD_CASE_DATA = load_case_data
            
            # Add to static list
            AnalysisControl.PDelta.data.append(self)
            
            # Automatically execute the data when instance is created
            self._execute()
        
        def _execute(self):
            """
            Automatically sends the P-Delta Analysis Control to the analysis system when created
            """
            json_data = {"Assign": {}}
            
            # Convert load case data to required format
            pdel_cases = []
            for case_name, factor in self.LOAD_CASE_DATA:
                pdel_cases.append({
                    "LCNAME": case_name,
                    "FACTOR": factor
                })
            
            control_data = {
                "ITER": self.ITER,
                "TOL": self.TOL,
                "PDEL_CASES": pdel_cases
            }
            
            json_data["Assign"][str(self.ID)] = control_data
            
            MidasAPI("PUT", "/db/pdel", json_data)

    class Buckling:
            """Create Buckling Analysis Control Object in Python"""
            data = []
            
            def __init__(self, 
                        mode_num: int = None,
                        opt_positive: bool = True,
                        load_factor_from: float = 0,
                        load_factor_to: float = 0,
                        opt_sturm_seq: bool = False,
                        opt_consider_axial_only: bool = False,
                        load_case_data: list = None):
                """
                Buckling Analysis Control constructor for eigenvalue buckling analysis settings.
                
                Parameters:
                    mode_num: Number of Modes (required)
                    
                    opt_positive: Load Factor Range Type (default True)
                    
                    load_factor_from: Search From (default 0)
                        - Lower bound for load factor search range
                        - Only used when opt_positive is False (Search mode)

                    load_factor_to: Search To (default 0)
                        - Upper bound for load factor search range
                        - Only used when opt_positive is False (Search mode)

                    opt_sturm_seq: Check Sturm Sequence (default False)
                    
                    opt_consider_axial_only: Frame Geometric Stiffness Option (default False)
                    
                    load_case_data: Load Cases with Scale Factors and Types (required)
                        - List of load cases with their scale factors and load types
                        - Format: [["LC1", factor1, load_type1], ["LC2", factor2, load_type2], ...]
                        - Load case name (string), scale factor (number), load type (integer)
                        - Load Type: 0=Variable, 1=Constant
                
                
                Examples:
                    Buckling(
                        mode_num=8, opt_positive=False, 
                        load_factor_from=-2.0, load_factor_to=5.0,
                        opt_consider_axial_only=True,
                        load_case_data=[["Gravity", 1.0, 1], ["Lateral", 1.0, 0]]
                    )
                """
                
                # Validate required parameters
                if mode_num is None:
                    raise ValueError("mode_num (Number of Modes) is required")
                if load_case_data is None or len(load_case_data) == 0:
                    raise ValueError("load_case_data (Load Cases) is required")
                
                # Validate load case data format
                for i, case in enumerate(load_case_data):
                    if not isinstance(case, list) or len(case) != 3:
                        raise ValueError(f"load_case_data[{i}] must be a list with 3 elements [name, factor, load_type]")
                    if not isinstance(case[0], str):
                        raise ValueError(f"load_case_data[{i}][0] (load case name) must be a string")
                    if not isinstance(case[1], (int, float)):
                        raise ValueError(f"load_case_data[{i}][1] (scale factor) must be a number")
                    if not isinstance(case[2], int) or case[2] not in [0, 1]:
                        raise ValueError(f"load_case_data[{i}][2] (load type) must be 0 (Variable) or 1 (Constant)")
                
                # ID is always 1
                self.ID = 1
                
                # Set parameters
                self.MODE_NUM = mode_num
                self.OPT_POSITIVE = opt_positive
                self.LOAD_FACTOR_FROM = load_factor_from
                self.LOAD_FACTOR_TO = load_factor_to
                self.OPT_STURM_SEQ = opt_sturm_seq
                self.OPT_CONSIDER_AXIAL_ONLY = opt_consider_axial_only
                self.LOAD_CASE_DATA = load_case_data
                
                # Add to static list
                AnalysisControl.Buckling.data.append(self)
                
                # Automatically execute the data when instance is created
                self._execute()
            
            def _execute(self):
                """
                Automatically sends the Buckling Analysis Control to the analysis system when created
                """
                json_data = {"Assign": {}}
                
                # Convert load case data to required format
                items = []
                for case_name, factor, load_type in self.LOAD_CASE_DATA:
                    items.append({
                        "LCNAME": case_name,
                        "FACTOR": factor,
                        "LOAD_TYPE": load_type
                    })
                
                control_data = {
                    "MODE_NUM": self.MODE_NUM,
                    "OPT_POSITIVE": self.OPT_POSITIVE,
                    "OPT_CONSIDER_AXIAL_ONLY": self.OPT_CONSIDER_AXIAL_ONLY,
                    "LOAD_FACTOR_FROM": self.LOAD_FACTOR_FROM,
                    "LOAD_FACTOR_TO": self.LOAD_FACTOR_TO,
                    "OPT_STURM_SEQ": self.OPT_STURM_SEQ,
                    "ITEMS": items
                }
                
                json_data["Assign"][str(self.ID)] = control_data
                
                MidasAPI("PUT", "/db/buck", json_data)


    class EigenValue:
        """Create Eigen Vector Analysis Control Object in Python"""
        data = []
        
        def __init__(self, 
                    analysis_type: str = None,
                    # EIGEN specific parameters
                    ifreq: int = 1,
                    iiter: int = 20,
                    idim: int = 1,
                    tol: float = 0,
                    # LANCZOS specific parameters 
                    frequency_range: list = None,  
                    bstrum: bool = False,
                    bminmax: bool = None,
                    frmin: float = None,
                    frmax: float = None,
                    # RITZ specific parameters 
                    bincnl: bool = False,
                    ignum: int = None,
                    load_vector: list = None,  
                    vritz: list = None):
            """
            Eigen Vector Analysis Control 
            
            Parameters:
                analysis_type: Type of Analysis (required)
                    - "EIGEN": Subspace Iteration
                    - "LANCZOS": Lanczos
                    - "RITZ": Ritz Vectors
                
                # For EIGEN:
                ifreq: Number of Frequencies (required for EIGEN)
                iiter: Number of Iterations (required for EIGEN)
                idim: Subspace Dimension (default 0, optional for EIGEN)
                tol: Convergence Tolerance (default 0, optional for EIGEN)
                
                # For LANCZOS :
                frequency_range: Frequency Range [frmin, frmax] (optional for LANCZOS)
                    - If provided, automatically sets bMINMAX=True
                    - Format: [min_freq, max_freq]
                bstrum: Sturm Sequence Check (default False, optional for LANCZOS)
                
                # For RITZ type only:
                bincnl: Include GL-link Force Vectors (default False, optional for RITZ)
                ignum: Number of Generations for Each GL-link Force (required for RITZ)
                load_vector: Load Cases in simple format (required for RITZ)
                    - Format: [["case_or_acc", nog], ...]
                    - For ground acceleration: ["ACCX"/"ACCY"/"ACCZ", nog]
                    - For load case: ["case_name", nog]
            
            Examples:
                # EIGEN analysis
                EigenValue(
                    analysis_type="EIGEN",
                    ifreq=10,
                    iiter=20,
                    idim=1,
                    tol=1e-10
                )
                
                # LANCZOS analysis
                EigenValue(
                    analysis_type="LANCZOS",
                    ifreq=15,
                    frequency_range=[0, 1600],  # Automatically sets bMINMAX=True
                    bstrum=True
                )
                
                # RITZ analysis
                EigenValue(
                    analysis_type="RITZ",
                    bincnl=False,
                    ignum=1,
                    load_vector=[["DL", 1], ["ACCX", 1]]  # Simple format
                )
            """
            
            # Validate required parameters
            if analysis_type is None:
                raise ValueError("analysis_type is required")
            if analysis_type not in ["EIGEN", "LANCZOS", "RITZ"]:
                raise ValueError("analysis_type must be 'EIGEN', 'LANCZOS', or 'RITZ'")
            
            # Validate type-specific required parameters
            if analysis_type in ["EIGEN"]:
                if ifreq is None:
                    raise ValueError("ifreq (Number of Frequencies) is required for EIGEN")
                if iiter is None:
                    raise ValueError("iiter (Number of Iterations) is required for EIGEN")
            
            # Handle LANCZOS parameters
            if analysis_type == "LANCZOS":
                # Handle new frequency_range format
                if frequency_range is not None:
                    if not isinstance(frequency_range, list) or len(frequency_range) != 2:
                        raise ValueError("frequency_range must be a list with exactly 2 elements [frmin, frmax]")
                    if frequency_range[0] >= frequency_range[1]:
                        raise ValueError("frmin must be less than frmax in frequency_range")
                    
                    # Automatically set parameters
                    bminmax = True
                    frmin = frequency_range[0]
                    frmax = frequency_range[1]
                else:
                    # Use legacy parameters or defaults
                    if bminmax is None:
                        bminmax = False
                    if bminmax and (frmin is None or frmax is None):
                        raise ValueError("frmin and frmax are required when bminmax is True for LANCZOS")
                    if frmin is not None and frmax is not None and frmin >= frmax:
                        raise ValueError("frmin must be less than frmax")
            
            # Handle RITZ parameters
            if analysis_type == "RITZ":
                if ignum is None:
                    raise ValueError("ignum (Number of Generations) is required for RITZ")
                
                # Handle new load_vector format
                if load_vector is not None:
                    if not isinstance(load_vector, list) or len(load_vector) == 0:
                        raise ValueError("load_vector must be a non-empty list")
                    
                    # Convert load_vector to vritz format
                    vritz = []
                    ground_acc_types = ["ACCX", "ACCY", "ACCZ"]
                    
                    for i, item in enumerate(load_vector):
                        if not isinstance(item, list) or len(item) != 2:
                            raise ValueError(f"load_vector[{i}] must be a list with exactly 2 elements [name, nog]")
                        
                        name, nog = item
                        if not isinstance(name, str):
                            raise ValueError(f"load_vector[{i}][0] (name) must be a string")
                        if not isinstance(nog, int) or nog <= 0:
                            raise ValueError(f"load_vector[{i}][1] (nog) must be a positive integer")
                        
                        # Determine if it's ground acceleration or case
                        if name in ground_acc_types:
                            vritz.append({
                                "KIND": "GROUND",
                                "GROUND": name,
                                "iNOG": nog
                            })
                        else:
                            vritz.append({
                                "KIND": "CASE",
                                "CASE": name,
                                "iNOG": nog
                            })
                
                # Use legacy vritz if provided and load_vector is not
                elif vritz is not None:
                    if not isinstance(vritz, list) or len(vritz) == 0:
                        raise ValueError("vritz (Load Cases) must be a non-empty list")
                    
                    # Validate legacy vritz format
                    for i, case in enumerate(vritz):
                        if not isinstance(case, dict):
                            raise ValueError(f"vritz[{i}] must be a dictionary")
                        if "KIND" not in case:
                            raise ValueError(f"vritz[{i}] must have 'KIND' key")
                        if case["KIND"] not in ["CASE", "GROUND"]:
                            raise ValueError(f"vritz[{i}]['KIND'] must be 'CASE' or 'GROUND'")
                        
                        if case["KIND"] == "GROUND":
                            if "GROUND" not in case or case["GROUND"] not in ["ACCX", "ACCY", "ACCZ"]:
                                raise ValueError(f"vritz[{i}] with KIND='GROUND' must have GROUND='ACCX'/'ACCY'/'ACCZ'")
                        elif case["KIND"] == "CASE":
                            if "CASE" not in case:
                                raise ValueError(f"vritz[{i}] with KIND='CASE' must have 'CASE' key")
                        
                        if "iNOG" not in case:
                            raise ValueError(f"vritz[{i}] must have 'iNOG' key")
                else:
                    raise ValueError("Either load_vector or vritz is required for RITZ analysis")
            
            # ID is always 1
            self.ID = 1
            
            # Set parameters
            self.TYPE = analysis_type
            self.iFREQ = ifreq
            self.iITER = iiter
            self.iDIM = idim
            self.TOL = tol
            self.bMINMAX = bminmax
            self.FRMIN = frmin
            self.FRMAX = frmax
            self.bSTRUM = bstrum
            self.bINCNL = bincnl
            self.iGNUM = ignum
            self.vRITZ = vritz
            
            # Add to static list
            AnalysisControl.EigenValue.data.append(self)
            
            # Automatically execute the data when instance is created
            self._execute()
        
        def _execute(self):
            """
            Automatically sends the Eigen Vector Analysis Control to the analysis system when created
            """
            json_data = {"Assign": {}}
            
            control_data = {"TYPE": self.TYPE}
            
            if self.TYPE in ["EIGEN", "LANCZOS"]:
                control_data.update({
                    "iFREQ": self.iFREQ,
                    "iITER": self.iITER,
                    "iDIM": self.iDIM,
                    "TOL": self.TOL
                })
                
                if self.TYPE == "LANCZOS":
                    control_data.update({
                        "bMINMAX": self.bMINMAX,
                        "FRMIN": self.FRMIN,
                        "FRMAX": self.FRMAX,
                        "bSTRUM": self.bSTRUM
                    })
            
            elif self.TYPE == "RITZ":
                control_data.update({
                    "bINCNL": self.bINCNL,
                    "iGNUM": self.iGNUM,
                    "vRITZ": self.vRITZ
                })
            
            json_data["Assign"][str(self.ID)] = control_data
            
            MidasAPI("PUT", "/db/eigv", json_data)

    class Settlement:
        
        data = []
        
        def __init__(self, 
                    concurrent_calc: bool = True,
                    concurrent_link: bool = True):
            """
            Settlement Analysis Control constructor for settlement analysis settings.
            
            Parameters:
                concurrent_calc: Plate Concurrent Force (default True, optional)
                    - Active: true
                    - Inactive: false
                
                concurrent_link: Elastic / General Links Concurrent Force (default True, optional)
                    - Active: true
                    - Inactive: false
            
            Examples:
                # with both Optional value
                Settlement()

                #without Optional value
                Settlement(
                    concurrent_calc=True,
                    concurrent_link=False
                )      

            """
            
            # ID is always 1
            self.ID = 1
            
            # Set parameters
            self.CONCURRENT_CALC = concurrent_calc
            self.CONCURRENT_LINK = concurrent_link
            
            # Add to static list
            AnalysisControl.Settlement.data.append(self)
            
            # Automatically execute the data when instance is created
            self._execute()
        
        def _execute(self):
            """
            Automatically sends the Settlement Analysis Control to the analysis system when created
            """
            json_data = {"Assign": {}}
            
            control_data = {
                "CONCURRENT_CALC": self.CONCURRENT_CALC,
                "CONCURRENT_LINK": self.CONCURRENT_LINK
            }
            
            json_data["Assign"][str(self.ID)] = control_data
            
            MidasAPI("PUT", "/db/smct", json_data)



from ._mapi import MidasAPI
# from ._node import *
from ._group import Group
from ._load import Load_Case

def convList(item):
    if type(item) != list:
        return [item]
    else:
        return item
    
class Temperature:

    @classmethod
    def create(cls):
        """Creates Temperature elements in MIDAS Civil NX"""
        if cls.System.temps: cls.System.create()
        if cls.Element.temps: cls.Element.create()
        if cls.Gradient.temps: cls.Gradient.create()
        if cls.Nodal.temps: cls.Nodal.create()
        if cls.BeamSection.temps: cls.BeamSection.create()        

    @classmethod
    def delete(cls):
        """Deletes Temperature elements from MIDAS Civil NX and Python"""
        cls.System.delete()
        cls.Element.delete()
        cls.Gradient.delete()
        cls.Nodal.delete()
        cls.BeamSection.delete()
        
    @classmethod
    def clear(cls):
        """Deletes Temperature elements from MIDAS Civil NX and Python"""
        cls.System.clear()
        cls.Element.clear()
        cls.Gradient.clear()
        cls.Nodal.clear()
        cls.BeamSection.clear()

    @classmethod
    def sync(cls):
        """Sync Temperature elements from MIDAS Civil NX to Python"""
        cls.System.sync()
        cls.Element.sync()
        cls.Gradient.sync()
        cls.Nodal.sync()
        cls.BeamSection.sync()
                
    # --------------------------------------------------------------------------------------------------
    # System Temperature
    # --------------------------------------------------------------------------------------------------
    class System:
        """
        Create System Temperature Object in Python
        
        Parameters:
            temperature (float): Temperature value
            lcname (str): Load case name
            group (str): Load group name (default "")
            id (int): System ID (optional)
        
        Example:
            Temperature.System(12.5, "Temp(+)", "LoadGroup1", 1)
        """
        temps = []
        
        def __init__(self, temperature, lcname, group="", id=None):
            chk = 0
            for i in Load_Case.cases:
                if lcname in i.NAME: chk = 1
            if chk == 0: Load_Case("T", lcname)

            if group:
                chk = 0
                try:
                    a = [v['NAME'] for v in Group.Load.json()["Assign"].values()]
                    if group in a:
                        chk = 1
                except:
                    pass
                if chk == 0:
                    Group.Load(group)
            
            self.TEMPER = temperature
            self.LCNAME = lcname
            self.GROUP_NAME = group
            
            if id is None:
                self.ID = len(Temperature.System.temps) + 1
            else:
                self.ID = id
            
            Temperature.System.temps.append(self)

        @classmethod
        def json(cls):
            """Creates JSON from System Temperature objects defined in Python"""
            json_data = {"Assign": {}}
            for temp in cls.temps:
                json_data["Assign"][str(temp.ID)] = {
                    "TEMPER": temp.TEMPER,
                    "LCNAME": temp.LCNAME,
                    "GROUP_NAME": temp.GROUP_NAME
                }
            return json_data

        @staticmethod
        def create():
            """Creates System Temperatures in MIDAS Civil NX"""
            MidasAPI("PUT", "/db/stmp", Temperature.System.json())

        @staticmethod
        def get():
            """Get the JSON of System Temperatures from MIDAS Civil NX"""
            return MidasAPI("GET", "/db/stmp")

        @staticmethod
        def sync():
            """Sync System Temperatures from MIDAS Civil NX to Python"""
            Temperature.System.temps = []
            a = Temperature.System.get()
            
            if a and 'STMP' in a:
                temp_data = a.get('STMP', {})
                for temp_id, temp_info in temp_data.items():
                    Temperature.System(
                        temp_info.get('TEMPER', 0),
                        temp_info.get('LCNAME', ''),
                        temp_info.get('GROUP_NAME', ''),
                        int(temp_id)
                    )

        @staticmethod
        def delete():
            """Delete System Temperatures from MIDAS Civil NX and Python"""
            Temperature.System.clear()
            return MidasAPI("DELETE", "/db/stmp")

        @staticmethod
        def clear():
            """Delete System Temperatures from Python"""
            Temperature.System.temps = []

    # --------------------------------------------------------------------------------------------------
    # Element Temperature
    # --------------------------------------------------------------------------------------------------
    class Element:
        """
        Create Element Temperature Object in Python
        
        Parameters:
            element (int): Element ID
            temperature (float): Temperature value
            lcname (str): Load case name
            group (str): Load group name (default "")
            id (int): Temperature ID (optional)
        
        Example:
            Temperature.Element(1, 35, "Temp(+)", "", 1)
        """
        temps = []
        
        def __init__(self, element, temperature, lcname, group="", id=None):

            chk = 0
            for i in Load_Case.cases:
                if lcname in i.NAME: chk = 1
            if chk == 0: Load_Case("T", lcname)

            if group:
                chk = 0
                try:
                    a = [v['NAME'] for v in Group.Load.json()["Assign"].values()]
                    if group in a:
                        chk = 1
                except:
                    pass
                if chk == 0:
                    Group.Load(group)
            
            self.ELEMENT = element
            self.TEMP = temperature
            self.LCNAME = lcname
            self.GROUP_NAME = group
            
            if id is None:
                existing_ids = []
                for temp in Temperature.Element.temps:
                    if temp.ELEMENT == element:
                        existing_ids.extend([item.get('ID', 0) for item in temp.ITEMS if hasattr(temp, 'ITEMS')])
                self.ID = max(existing_ids, default=0) + 1
            else:
                self.ID = id
            
            existing_temp = None
            for temp in Temperature.Element.temps:
                if temp.ELEMENT == element:
                    existing_temp = temp
                    break
            
            item_data = {
                "ID": self.ID, "LCNAME": self.LCNAME,
                "GROUP_NAME": self.GROUP_NAME, "TEMP": self.TEMP
            }

            if existing_temp:
                if not hasattr(existing_temp, 'ITEMS'):
                    existing_temp.ITEMS = []
                existing_temp.ITEMS.append(item_data)
            else:
                self.ITEMS = [item_data]
                Temperature.Element.temps.append(self)

        @classmethod
        def json(cls):
            """Creates JSON from Element Temperature objects defined in Python"""
            json_data = {"Assign": {}}
            for temp in cls.temps:
                json_data["Assign"][str(temp.ELEMENT)] = {"ITEMS": temp.ITEMS}
            return json_data

        @staticmethod
        def create():
            """Creates Element Temperatures in MIDAS Civil NX"""
            MidasAPI("PUT", "/db/etmp", Temperature.Element.json())

        @staticmethod
        def get():
            """Get the JSON of Element Temperatures from MIDAS Civil NX"""
            return MidasAPI("GET", "/db/etmp")

        @staticmethod
        def sync():
            """Sync Element Temperatures from MIDAS Civil NX to Python"""
            Temperature.Element.temps = []
            a = Temperature.Element.get()
            
            if a and 'ETMP' in a:
                temp_data = a.get('ETMP', {})
                for element_id, element_data in temp_data.items():
                    element_obj = type('obj', (object,), {
                        'ELEMENT': int(element_id),
                        'ITEMS': element_data.get('ITEMS', [])
                    })()
                    Temperature.Element.temps.append(element_obj)

        @staticmethod
        def delete():
            """Delete Element Temperatures from MIDAS Civil NX and Python"""
            Temperature.Element.clear()
            return MidasAPI("DELETE", "/db/etmp")
        
        @staticmethod
        def clear():
            """Delete Element Temperatures from MIDAS Civil NX and Python"""
            Temperature.Element.temps = []

    # --------------------------------------------------------------------------------------------------
    # Temperature Gradient
    # --------------------------------------------------------------------------------------------------
    class Gradient:
        """
        Create Temperature Gradient Object in Python for Beam and Plate elements.
        
        Parameters:
            element (int): Element ID to apply the gradient.
            type (str): Element type, either 'Beam' or 'Plate'.
            lcname (str): Load Case Name (must exist in the model).
            tz (float): Temperature difference in the local z-direction (T2z - T1z).
            group (str, optional): Load Group Name. Defaults to "".
            id (int, optional): Gradient ID. Auto-assigned if not provided.
            hz (float, optional): Gradient value for local z-dir. If omitted, section default is used.
            ty (float, optional): Temp. diff. in local y-dir (T2y - T1y). **Required for 'Beam' type.**
            hy (float, optional): Gradient value for local y-dir. If omitted, section default is used.
        
        Example for Beam (providing gradient values):
            Temperature.Gradient(element=2, type='Beam', lcname='Temp(-)', tz=10, ty=-10, hz=1.2, hy=0.5)
        
        Example for Beam (using section defaults):
            Temperature.Gradient(element=2, type='Beam', lcname='Temp(+)', tz=10, ty=-10)
        
        Example for Plate (providing gradient value):
            Temperature.Gradient(element=21, type='Plate', lcname='Temp(-)', tz=10, hz=0.2)
        """
        temps = []

        def __init__(self, element, type, lcname, tz, group="", hz=None, ty=0, hy=None,id=None):

            chk = 0
            for i in Load_Case.cases:
                if lcname in i.NAME: chk = 1
            if chk == 0: Load_Case("T", lcname)
            
            if group:
                chk = 0
                try:
                    a = [v['NAME'] for v in Group.Load.json()["Assign"].values()]
                    if group in a:
                        chk = 1
                except:
                    pass
                if chk == 0:
                    Group.Load(group)
            
            self.ELEMENT = element
            
            if id is None:
                existing_ids = []
                for temp in Temperature.Gradient.temps:
                    if temp.ELEMENT == element:
                        existing_ids.extend([item.get('ID', 0) for item in temp.ITEMS if hasattr(temp, 'ITEMS')])
                self.ID = max(existing_ids, default=0) + 1
            else:
                self.ID = id

            use_hz = (hz is None)
            use_hy = (hy is None)

            item_data = {
                "ID": self.ID,
                "LCNAME": lcname,
                "GROUP_NAME": group,
                "TZ": tz,
                "USE_HZ": use_hz,
            }

            if not use_hz:
                item_data["HZ"] = hz

            if type.lower() == 'beam':
                item_data["TYPE"] = 1
                item_data["TY"] = ty
                item_data["USE_HY"] = use_hy
                if not use_hy:
                    item_data["HY"] = hy
            elif type.lower() == 'plate':
                item_data["TYPE"] = 2
            else:
                raise ValueError("Element type for Gradient must be 'Beam' or 'Plate'.")

            existing_temp = None
            for temp in Temperature.Gradient.temps:
                if temp.ELEMENT == element:
                    existing_temp = temp
                    break

            if existing_temp:
                if not hasattr(existing_temp, 'ITEMS'):
                    existing_temp.ITEMS = []
                existing_temp.ITEMS.append(item_data)
            else:
                self.ITEMS = [item_data]
                Temperature.Gradient.temps.append(self)
        
        @classmethod
        def json(cls):
            """Creates JSON from Temperature Gradient objects defined in Python"""
            json_data = {"Assign": {}}
            for temp in cls.temps:
                json_data["Assign"][str(temp.ELEMENT)] = {"ITEMS": temp.ITEMS}
            return json_data

        @staticmethod
        def create():
            """Creates Temperature Gradients in MIDAS Civil NX"""
            MidasAPI("PUT", "/db/gtmp", Temperature.Gradient.json())

        @staticmethod
        def get():
            """Get the JSON of Temperature Gradients from MIDAS Civil NX"""
            return MidasAPI("GET", "/db/gtmp")

        @staticmethod
        def sync():
            """Sync Temperature Gradients from MIDAS Civil NX to Python"""
            Temperature.Gradient.temps = []
            a = Temperature.Gradient.get()
            
            if a and 'GTMP' in a:
                temp_data = a.get('GTMP', {})
                for element_id, element_data in temp_data.items():
                    element_obj = type('obj', (object,), {
                        'ELEMENT': int(element_id),
                        'ITEMS': element_data.get('ITEMS', [])
                    })()
                    Temperature.Gradient.temps.append(element_obj)

        @staticmethod
        def delete():
            """Delete Temperature Gradients from MIDAS Civil NX and Python"""
            Temperature.Gradient.clear()
            return MidasAPI("DELETE", "/db/gtmp")
        
        @staticmethod
        def clear():
            """Delete Temperature Gradients from MIDAS Civil NX and Python"""
            Temperature.Gradient.temps = []

    # --------------------------------------------------------------------------------------------------
    # Nodal Temperature
    # --------------------------------------------------------------------------------------------------
    class Nodal:
        """
        Create Nodal Temperature  
        
        Parameters:
            node (int): Node ID
            temperature (float): Temperature value  
            lcname (str): Load case name **(Must exist in the model)**
            group (str): Load group name (default "")
            id (int): Temperature ID (optional)
        
        Example:
            Temperature.Nodal(6, 10, "Test")
        """
        temps = []
        
        def __init__(self, node, temperature, lcname, group="", id=None):

            chk = 0
            for i in Load_Case.cases:
                if lcname in i.NAME: chk = 1
            if chk == 0: Load_Case("T", lcname)

            if group:
                chk = 0
                try:
                    a = [v['NAME'] for v in Group.Load.json()["Assign"].values()]
                    if group in a:
                        chk = 1
                except:
                    pass
                if chk == 0:
                    Group.Load(group)
            
            self.NODE = node
            self.TEMPER = temperature
            self.LCNAME = lcname
            self.GROUP_NAME = group
            
            if id is None:
                existing_ids = []
                for temp in Temperature.Nodal.temps:
                    if temp.NODE == node:
                        existing_ids.extend([item.get('ID', 0) for item in temp.ITEMS if hasattr(temp, 'ITEMS')])
                self.ID = max(existing_ids, default=0) + 1
            else:
                self.ID = id
            
            existing_temp = None
            for temp in Temperature.Nodal.temps:
                if temp.NODE == node:
                    existing_temp = temp
                    break
            
            item_data = {
                "ID": self.ID, "LCNAME": self.LCNAME,
                "GROUP_NAME": self.GROUP_NAME, "TEMPER": self.TEMPER
            }

            if existing_temp:
                if not hasattr(existing_temp, 'ITEMS'):
                    existing_temp.ITEMS = []
                existing_temp.ITEMS.append(item_data)
            else:
                self.ITEMS = [item_data]
                Temperature.Nodal.temps.append(self)

        @classmethod
        def json(cls):
            """Creates JSON with 'Assign' key from Nodal Temperature objects defined in Python"""
            json_data = {"Assign": {}}
            for temp in cls.temps:
                json_data["Assign"][str(temp.NODE)] = {"ITEMS": temp.ITEMS}
            return json_data

        @staticmethod
        def create():
            """Creates Nodal Temperatures in MIDAS Civil NX"""
            MidasAPI("PUT", "/db/ntmp", Temperature.Nodal.json())

        @staticmethod
        def get():
            """Get the JSON of Nodal Temperatures from MIDAS Civil NX"""
            return MidasAPI("GET", "/db/ntmp")

        @staticmethod
        def sync():
            """Sync Nodal Temperatures from MIDAS Civil NX to Python"""
            Temperature.Nodal.temps = []
            a = Temperature.Nodal.get()
            
            if a and 'NTMP' in a:
                temp_data = a.get('NTMP', {})
                for node_id, node_data in temp_data.items():
                    node_obj = type('obj', (object,), {
                        'NODE': int(node_id),
                        'ITEMS': node_data.get('ITEMS', [])
                    })()
                    Temperature.Nodal.temps.append(node_obj)

        @staticmethod
        def delete():
            """Delete Nodal Temperatures from MIDAS Civil NX and Python"""
            Temperature.Nodal.clear()
            return MidasAPI("DELETE", "/db/ntmp")
        
        @staticmethod
        def clear():
            """Delete Nodal Temperatures from MIDAS Civil NX and Python"""
            Temperature.Nodal.temps = []


    # --------------------------------------------------------------------------------------------------
    # Beam Section Temperature
    # --------------------------------------------------------------------------------------------------
    class BeamSection:
        """
        Create Beam Section Temperature Object in Python.
        
        Parameters:
            element (int): Element ID to apply the load.
            lcname (str): Load Case Name (must exist in the model).
            section_type (str, optional): 'General' or 'PSC'. Defaults to 'General'.
            type (str, optional): 'Element' or 'Input'. Defaults to 'Element'.
            group (str, optional): Load Group Name.
            id (int, optional): Load ID.
            dir (str, optional): Direction, 'LY' or 'LZ'. Defaults to 'LZ'.
            ref_pos (str, optional): Reference Position, 'Centroid', 'Top', or 'Bot'. Defaults to 'Centroid'.
            val_b (float, optional): B Value.
            val_h1 (float, optional): H1 Value.
            val_h2 (float, optional): H2 Value.
            val_t1 (float, optional): T1 Value.
            val_t2 (float, optional): T2 Value.
            elast (float, optional): Modulus of Elasticity (required for 'Input' type).
            thermal (float, optional): Thermal Coefficient (required for 'Input' type).
            psc_ref (int, optional): Reference for PSC, 0 for Top, 1 for Bottom.
            psc_opt_b (int, optional): B-Type option for PSC. (0 for Section type)
            psc_opt_h1 (int, optional): H1-Type option for PSC. (0 - Z1 , 1- Z2 ,2 - Z2)
            psc_opt_h2 (int, optional): H2-Type option for PSC.  (0 - Z1 , 1- Z2 ,2 - Z2)
        """
        temps = []

        def __init__(self, element, lcname, section_type='General', type='Element', group="", 
                    dir='LZ', ref_pos='Centroid', val_b=0, val_h1=0, val_h2=0, val_t1=0, val_t2=0,
                    elast=None, thermal=None, psc_ref=0, psc_opt_b=1, psc_opt_h1=3, psc_opt_h2=3, id=None):

            # Validate required parameters for Input type
            if type.upper() == 'INPUT':
                if elast is None or thermal is None:
                    raise ValueError("For 'Input' type, both 'elast' and 'thermal' parameters are required.")

            # Handle load group creation

            chk = 0
            for i in Load_Case.cases:
                if lcname in i.NAME: chk = 1
            if chk == 0: Load_Case("T", lcname)
            
            if group:
                chk = 0
                try:
                    a = [v['NAME'] for v in Group.Load.json()["Assign"].values()]
                    if group in a:
                        chk = 1
                except:
                    pass
                if chk == 0:
                    Group.Load(group)
            
            self.ELEMENT = element

            # Auto-assign ID if not provided
            if id is None:
                existing_ids = []
                for temp in Temperature.BeamSection.temps:
                    if temp.ELEMENT == element:
                        existing_ids.extend([item.get('ID', 0) for item in temp.ITEMS if hasattr(temp, 'ITEMS')])
                self.ID = max(existing_ids, default=0) + 1
            else:
                self.ID = id
            
            # Construct the nested dictionary for vSECTTMP
            vsec_item = {
                "TYPE": type.upper(),
                "VAL_B": val_b,
                "VAL_H1": val_h1,
                "VAL_H2": val_h2,
                "VAL_T1": val_t1,
                "VAL_T2": val_t2,
            }

            is_psc = section_type.lower() == 'psc'
            
            # Add material properties for Input type
            if type.upper() == 'INPUT':
                vsec_item["ELAST"] = elast
                vsec_item["THERMAL"] = thermal
            
            # Add PSC-specific parameters
            if is_psc:
                vsec_item["REF"] = psc_ref
                vsec_item["OPT_B"] = psc_opt_b
                vsec_item["OPT_H1"] = psc_opt_h1
                vsec_item["OPT_H2"] = psc_opt_h2

            # Construct the main item dictionary
            item_data = {
                "ID": self.ID,
                "LCNAME": lcname,
                "GROUP_NAME": group,
                "DIR": dir,
                "REF": ref_pos,
                "NUM": 1,
                "bPSC": is_psc,
                "vSECTTMP": [vsec_item]
            }

            # Check if an object for this element already exists
            existing_temp = None
            for temp in Temperature.BeamSection.temps:
                if temp.ELEMENT == element:
                    existing_temp = temp
                    break

            if existing_temp:
                if not hasattr(existing_temp, 'ITEMS'):
                    existing_temp.ITEMS = []
                existing_temp.ITEMS.append(item_data)
            else:
                self.ITEMS = [item_data]
                Temperature.BeamSection.temps.append(self)
        
        @classmethod
        def json(cls):
            """Creates JSON from Beam Section Temperature objects defined in Python"""
            json_data = {"Assign": {}}
            for temp in cls.temps:
                json_data["Assign"][str(temp.ELEMENT)] = {"ITEMS": temp.ITEMS}
            return json_data

        @staticmethod
        def create():
            """Creates Beam Section Temperatures in MIDAS Civil NX"""
            MidasAPI("PUT", "/db/btmp", Temperature.BeamSection.json())

        @staticmethod
        def get():
            """Get the JSON of Beam Section Temperatures from MIDAS Civil NX"""
            return MidasAPI("GET", "/db/btmp")

        @staticmethod
        def sync():
            """Sync Beam Section Temperatures from MIDAS Civil NX to Python"""
            Temperature.BeamSection.temps = []
            a = Temperature.BeamSection.get()
            
            if a and 'BTMP' in a:
                temp_data = a.get('BTMP', {})
                for element_id, element_data in temp_data.items():
                    element_obj = type('obj', (object,), {
                        'ELEMENT': int(element_id),
                        'ITEMS': element_data.get('ITEMS', [])
                    })()
                    Temperature.BeamSection.temps.append(element_obj)

        @staticmethod
        def delete():
            """Delete Beam Section Temperatures from MIDAS Civil NX and Python"""
            Temperature.BeamSection.clear()
            return MidasAPI("DELETE", "/db/btmp")
        
        @staticmethod
        def clear():
            """Delete Beam Section Temperatures from MIDAS Civil NX and Python"""
            Temperature.BeamSection.temps = []